# -*- coding: utf-8 -*-

"""
.. module:: activity_classes.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Set of CIM v1 activity package class definitions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""


def activity():
    """An abstract class used as the parent of MeasurementCampaigns, Projects, Experiments, and NumericalActivities.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': True,
        'properties': [
            ('funding_sources', 'str', '0.N'),
            ('projects', 'activity.project_type', '0.N'),
            ('rationales', 'str', '0.N'),
            ('responsible_parties', 'shared.responsible_party', '0.N'),
        ],
        'doc_strings': {
            'funding_sources': 'The entities that funded this activity.',
            'projects': 'The project(s) that this activity is associated with (ie: CMIP5, AMIP, etc).',
            'rationales': 'For what purpose is this activity being performed?',
            'responsible_parties': 'The point of contact(s) for this activity.This includes, among others, the principle investigator.',
        },
        'decodings': [
            ('funding_sources', 'child::cim:fundingSource'),
            ('projects', 'child::cim:project/@value'),
            ('rationales', 'child::cim:rationale'),
            ('responsible_parties', 'child::cim:responsibleParty'),
        ]
    }


def conformance():
    """A conformance class maps how a configured model component met a specific numerical requirement.  For example, for a double CO2 boundary condition, a model component might read a CO2 dataset in which CO2 has been doubled, or it might modify a parameterisation (presumably with a factor of two somewhere).  So, the conformance links a requirement to a DataSource (which can be either an actual DataObject or a property of a model component).  In some cases a model/simulation may _naturally_ conform to a requirement.  In this case there would be no reference to a DataSource but the conformant attribute would be true.  If something is purpopsefully non-conformant then the conformant attribute would be false.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('description', 'str', '0.1'),
            ('frequency', 'activity.frequency_type', '0.1'),
            ('is_conformant', 'bool', '1.1'),
            ('requirements', 'activity.numerical_requirement', '1.N'),
            ('sources', 'shared.data_source', '0.N'),
            ('type', 'activity.conformance_type', '0.1'),
        ],
        'doc_strings': {
            'is_conformant': 'Records whether or not this conformance satisfies the requirement.  A simulation should have at least one conformance mapping to every experimental requirement.  If a simulation satisfies the requirement - the usual case - then conformant should have a value of true.  If conformant is true but there is no reference to a source for the conformance, then we can assume that the simulation conforms to the requirement _naturally_, that is without having to modify code or inputs. If a simulation does not conform to a requirement then conformant should be set to false.',
            'requirements': 'Points to the NumericalRequirement that the simulation in question is conforming to.',
            'sources': 'Points to the DataSource used to conform to a particular Requirement.   This may be part of an activity::simulation or a software::component.  It can be either a DataObject or a SoftwareComponent or a ComponentProperty.  It could also be by using particular attributes of, say, a SoftwareComponent, but in that case the recommended practise is to reference the component and add appropriate text in the conformance description attribute.',
            'type': 'Describes the method that this simulation conforms to an experimental requirement (in case it is not specified by the change property of the reference to the source of this conformance)'
        },
        'decodings': [
            ('description', 'child::cim:description'),
            ('frequency', 'child::cim:frequency'),
            ('is_conformant', '@conformant'),
            ('requirements', 'child::cim:requirement/cim:requirement/cim:initialCondition', 'activity.initial_condition'),
            ('requirements', 'child::cim:requirement/cim:requirement/cim:boundaryCondition', 'activity.boundary_condition'),
            ('requirements', 'child::cim:requirement/cim:requirement/cim:lateralBoundaryCondition', 'activity.lateral_boundary_condition'),
            ('requirements', 'child::cim:requirement/cim:requirement/cim:spatioTemporalConstraint', 'activity.spatio_temporal_constraint'),
            ('requirements', 'child::cim:requirement/cim:requirement/cim:outputRequirement', 'activity.output_requirement'),
            ('requirements', 'child::cim:requirement/cim:reference', 'shared.doc_reference'),
            ('sources', 'child::cim:source/cim:source/cim:dataObject', 'data.data_object'),
            ('sources', 'child::cim:source/cim:source/cim:dataContent', 'data.data_content'),
            ('sources', 'child::cim:source/cim:source/cim:componentProperty', 'software.component_property'),
            ('sources', 'child::cim:source/cim:source/cim:softwareComponent', 'software.model_component'),
            ('sources', 'child::cim:source/cim:source/cim:softwareComponent', 'software.processor_component'),
            ('sources', 'child::cim:source/cim:source/cim:softwareComponent', 'software.statistical_model_component'),
            ('sources', 'child::cim:source/cim:reference', 'shared.doc_reference'),
            ('type', '@type'),
        ]
    }


def downscaling_simulation():
    """A simulation is the implementation of a numerical experiment.  A simulation can be made up of "child" simulations aggregated together to form a simulation composite.  The parent simulation can be made up of whole or partial child simulations, the simulation attributes need to be able to capture this.

    """
    return {
        'type': 'class',
        'base': 'activity.numerical_activity',
        'is_abstract': True,
        'is_document': True,
        'properties': [
            ('calendar', 'shared.calendar', '1.1'),
            ('inputs', 'software.coupling', '0.N'),
            ('outputs', 'data.data_object', '0.N'),
            ('downscaling_id', 'str', '0.1'),
            ('downscaled_from', 'shared.data_source', '1.1'),
            ('downscaling_type', 'activity.downscaling_type', '0.1')
        ],
        'doc_strings': {
            'inputs': 'Implemented as a mapping from a source to target; can be a forcing file, a boundary condition, etc.',
        },
        'decodings': [
            ('meta', 'self::cim:downscalingSimulation'),
            ('calendar', 'child::cim:calendar/cim:daily-360', 'shared.daily_360'),
            ('calendar', 'child::cim:calendar/cim:perpetualPeriod', 'shared.perpetual_period'),
            ('calendar', 'child::cim:calendar/cim:realCalendar', 'shared.real_calendar'),
            ('inputs', 'child::cim:input'),
            ('outputs', 'child::cim:output/cim:dataObject', 'data.data_object'),
            ('outputs', 'child::cim:output/cim:reference', 'shared.doc_reference'),
            ('downscaling_id', 'child::cim:downscalingID'),
            ('downscaled_from', 'child::cim:downscaledFrom/cim:downscaledFrom/cim:dataObject', 'data.data_object'),
            ('downscaled_from', 'child::cim:downscaledFrom/cim:downscaledFrom/cim:dataContent', 'data.data_content'),
            ('downscaled_from', 'child::cim:downscaledFrom/cim:downscaledFrom/cim:componentProperty', 'software.component_property'),
            ('downscaled_from', 'child::cim:downscaledFrom/cim:downscaledFrom/cim:softwareComponent', 'software.model_component'),
            ('downscaled_from', 'child::cim:downscaledFrom/cim:downscaledFrom/cim:softwareComponent', 'software.processor_component'),
            ('downscaled_from', 'child::cim:downscaledFrom/cim:downscaledFrom/cim:softwareComponent', 'software.statistical_model_component'),
            ('downscaled_from', 'child::cim:downscaledFrom/cim:reference', 'shared.doc_reference'),
            ('downscaling_type', 'self::cim:downscalingSimulation/@downscalingType'),
        ]
    }


def ensemble():
    """An ensemble is made up of two or more simulations which are to be compared against each other to create ensemble statistics. Ensemble members can differ in terms of initial conditions, physical parameterisation and the model used. An ensemble bundles together sets of ensembleMembers, all of which reference the same Simulation(Run) and include one or more changes.

    """
    return {
        'type': 'class',
        'base': 'activity.numerical_activity',
        'is_abstract': False,
        'is_document': True,
        'properties': [
            ('members', 'activity.ensemble_member', '1.N'),
            ('types', 'activity.ensemble_type', '1.N'),
            ('outputs', 'shared.data_source', '0.N')
        ],
        'doc_strings': {
            'outputs': 'Points to the DataSource used to conform to a particular Requirement.   This may be part of an activity::simulation or a software::component.  It can be either a DataObject or a SoftwareComponent or a ComponentProperty.  It could also be by using particular attributes of, say, a SoftwareComponent, but in that case the recommended practise is to reference the component and add appropriate text in the conformance description attribute.',
        },
        'decodings': [
            ('meta', 'self::cim:ensemble'),
            ('members', 'child::cim:ensembleMember'),
            ('types', 'child::cim:ensembleType/@value'),
            ('outputs', 'child::cim:output/cim:output/cim:dataObject', 'data.data_object'),
            ('outputs', 'child::cim:output/cim:output/cim:dataContent', 'data.data_content'),
            ('outputs', 'child::cim:output/cim:output/cim:componentProperty', 'software.component_property'),
            ('outputs', 'child::cim:output/cim:output/cim:softwareComponent', 'software.model_component'),
            ('outputs', 'child::cim:output/cim:output/cim:softwareComponent', 'software.processor_component'),
            ('outputs', 'child::cim:output/cim:output/cim:softwareComponent', 'software.statistical_model_component'),
            ('outputs', 'child::cim:output/cim:reference', 'shared.doc_reference')
        ]
    }


def ensemble_member():
    """A simulation is the implementation of a numerical experiment.  A simulation can be made up of "child" simulations aggregated together to form a "simulation composite".  The "parent" simulation can be made up of whole or partial child simulations, the simulation attributes need to be able to capture this.

    """
    return {
        'type': 'class',
        'base': 'activity.numerical_activity',
        'is_abstract': False,
        'properties': [
            ('ensemble', 'activity.ensemble', '0.1'),
            ('ensemble_ids', 'shared.standard_name', '0.N'),
            ('simulation', 'activity.simulation', '0.1')
        ],
        'decodings': [
            ('ensemble', 'child::cim:ensemble/cim:ensemble'),
            ('ensemble', 'child::cim:ensemble/cim:reference', 'shared.doc_reference'),
            ('ensemble_ids', 'child::cim:ensembleMemberID'),
            ('simulation', 'child::cim:ensemble/cim:simulation'),
            ('simulation', 'child::cim:simulation/cim:reference', 'shared.doc_reference'),
        ]
    }


def experiment():
    """An experiment might be an activity which is both observational and numerical in focus, for example, a measurement campaign and numerical experiments for an alpine experiment.  It is a place for the scientific description of the reason why an experiment was made.

    """
    return {
        'type': 'class',
        'base': 'activity.activity',
        'is_abstract': True,
        'properties': [
            ('measurement_campaigns', 'activity.measurement_campaign', '0.N'),
            ('requires', 'activity.numerical_activity', '0.N'),
            ('generates', 'str', '0.N'),
            ('supports', 'str', '0.N'),
        ],
    }


def experiment_relationship():
    """Contains a set of relationship types specific to a experiment document that can be used to describe its genealogy.

    """
    return {
        'type': 'class',
        'base': 'shared.relationship',
        'is_abstract': False,
        'properties': [
            ('target', 'activity.experiment_relationship_target', '1.1'),
            ('type', 'activity.experiment_relationship_type', '1.1'),
        ],
    }


def experiment_relationship_target():
    """Creates and returns instance of experiment_relationship_target class.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('numerical_experiment', 'activity.numerical_experiment', '0.1')
        ],
    }


def lateral_boundary_condition():
    """A boundary condition is a numerical requirement which looks like a variable imposed on the model evolution (i.e. it might - or might not - evolve with time, but is seen by the model at various times during its evolution) as opposed to an initial condition (at model time zero).

    """
    return {
        'type': 'class',
        'base': 'activity.numerical_requirement',
        'is_abstract': False,
        'constraints': [
            ('requirement_type', 'constant', 'lateralBoundaryCondition'),
        ]
    }


def measurement_campaign():
    """Creates and returns instance of measurement_campaign class.

    """
    return {
        'type': 'class',
        'base': 'activity.activity',
        'is_abstract': False,
        'properties': [
            ('duration', 'int', '1.1'),
        ],
    }


def numerical_activity():
    """Creates and returns instance of numerical_activity class.

    """
    return {
        'type': 'class',
        'base': 'activity.activity',
        'is_abstract': True,
        'properties': [
            ('description', 'str', '0.1'),
            ('long_name', 'str', '0.1'),
            ('short_name', 'str', '1.1'),
            ('supports', 'activity.experiment', '0.N'),
        ],
        'doc_strings': {
            'description': 'A free-text description of the experiment.',
            'long_name': 'The name of the experiment (that is recognized externally).',
            'short_name': 'The name of the experiment (that is used internally).'
        },
        'decodings': [
            ('description', 'child::cim:description'),
            ('long_name', 'child::cim:longName'),
            ('short_name', 'child::cim:shortName'),
            ('supports', 'child::cim:supports/cim:experiment'),
            ('supports', 'child::cim:supports/cim:reference', 'shared.doc_reference'),
        ]
    }


def numerical_experiment():
    """A numerical experiment may be generated by an experiment, in which case it is inSupportOf the experiment. But a numerical experiment may also exist as an activity in its own right (as it might be if it were needed for a MIP). Examples: AR4 individual experiments, AR5 individual experiments, RAPID THC experiments etc.

    """
    return {
        'type': 'class',
        'base': 'activity.experiment',
        'is_abstract': False,
        'is_document': True,
        'properties': [
            ('description', 'str', '0.1'),
            ('experiment_id', 'str', '0.1'),
            ('long_name', 'str', '0.1'),
            ('requirements', 'activity.numerical_requirement', '0.N'),
            ('short_name', 'str', '1.1'),
        ],
        'doc_strings': {
            'description': 'A free-text description of the experiment.',
            'experiment_id': 'An experiment ID takes the form <number>.<number>[-<letter>].',
            'long_name': 'The name of the experiment (that is recognized externally).',
            'requirements': 'The name of the experiment (that is used internally).',
            'short_name': 'The name of the experiment (that is used internally).',
        },
        'decodings': [
            ('meta', 'self::cim:numericalExperiment'),
            ('description', 'child::cim:description'),
            ('experiment_id', 'child::cim:experimentID'),
            ('long_name', 'child::cim:longName'),
            ('short_name', 'child::cim:shortName'),
            # Metafor CMIP5 questionnaire - standalone
            ('requirements', 'child::cim:numericalRequirement[@xsi:type="InitialCondition"]', 'activity.initial_condition'),
            ('requirements', 'child::cim:numericalRequirement[@xsi:type="BoundaryCondition"]', 'activity.boundary_condition'),
            ('requirements', 'child::cim:numericalRequirement[@xsi:type="LateralBoundaryCondition"]', 'activity.lateral_boundary_condition'),
            ('requirements', 'child::cim:numericalRequirement[@xsi:type="SpatioTemporalConstraint"]', 'activity.spatio_temporal_constraint'),
            ('requirements', 'child::cim:numericalRequirement[@xsi:type="OutputRequirement"]', 'activity.output_requirement'),
            # Metafor CMIP5 questionnaire - simulation bundle
            ('requirements', 'child::cim:numericalRequirement/cim:initialCondition', 'activity.initial_condition'),
            ('requirements', 'child::cim:numericalRequirement/cim:boundaryCondition', 'activity.boundary_condition'),
            ('requirements', 'child::cim:numericalRequirement/cim:lateralBoundaryCondition', 'activity.lateral_boundary_condition'),
            ('requirements', 'child::cim:numericalRequirement/cim:spatioTemporalConstraint', 'activity.spatio_temporal_constraint'),
            ('requirements', 'child::cim:numericalRequirement/cim:outputRequirement', 'activity.output_requirement'),
        ]
    }


def numerical_requirement():
    """A description of the requirements of particular experiments.  Numerical Requirements can be initial conditions, boundary conditions, or physical modificiations.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': True,
        'properties': [
            ('description', 'str', '0.1'),
            ('id', 'str', '0.1'),
            ('name', 'str', '1.1'),
            ('options', 'activity.numerical_requirement_option', '0.N'),
            ('requirement_type', 'str', '1.1'),
            ('source', 'shared.data_source', '0.1')
        ],
        'doc_strings': {
            'requirement_type': 'Type of reqirement to which the experiment must conform.',
        },
        'decodings': [
            ('description', 'child::cim:description'),
            ('id', 'child::cim:id'),
            ('name', 'child::cim:name'),
            ('options', 'child::cim:requirementOption'),
            ('source', 'child::cim:source/cim:source/cim:dataObject', 'data.data_object'),
            ('source', 'child::cim:source/cim:source/cim:dataContent', 'data.data_content'),
            ('source', 'child::cim:source/cim:source/cim:componentProperty', 'software.component_property'),
            ('source', 'child::cim:source/cim:source/cim:softwareComponent', 'software.model_component'),
            ('source', 'child::cim:source/cim:source/cim:softwareComponent', 'software.processor_component'),
            ('source', 'child::cim:source/cim:source/cim:softwareComponent', 'software.statistical_model_component'),
            ('source', 'child::cim:source/cim:reference', 'shared.doc_reference')
        ]
    }


def numerical_requirement_option():
    """A NumericalRequirement that is being used as a set of related requirements; For example if a requirement is to use 1 of 3 boundary conditions, then that "parent" requirement would have three "child" RequirmentOptions (each of one with the XOR optionRelationship).

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('description', 'str', '0.1'),
            ('id', 'str', '0.1'),
            ('name', 'str', '1.1'),
            ('sub_options', 'activity.numerical_requirement_option', '0.N'),
            ('relationship', 'str', '0.1'),
        ],
        'doc_strings': {
            'relationship': 'Describes how this optional (child) requirement is related to its sibling requirements.  For example, a NumericalRequirement could consist of a set of optional requirements each with an "OR" relationship meaning use this boundary condition _or_ that one.'
        },
        'decodings': [
            ('description', 'child::cim:description'),
            ('id', 'child::cim:id'),
            ('name', 'child::cim:name'),
            ('relationship', 'self::cim:requirementOption/@optionRelationship'),
            # Metafor CMIP5 questionnaire - simulation bundle
            ('description', 'child::cim:requirement/cim:requirement/cim:boundaryCondition/cim:description'),
            ('description', 'child::cim:requirement/cim:requirement/cim:initialCondition/cim:description'),
            ('description', 'child::cim:requirement/cim:requirement/cim:lateralBoundaryCondition/cim:description'),
            ('description', 'child::cim:requirement/cim:requirement/cim:outputRequirement/cim:description'),
            ('description', 'child::cim:requirement/cim:requirement/cim:spatioTemporalConstraint/cim:description'),
            # Metafor CMIP5 questionnaire - simulation bundle
            ('id', 'child::cim:requirement/cim:requirement/cim:boundaryCondition/cim:id'),
            ('id', 'child::cim:requirement/cim:requirement/cim:initialCondition/cim:id'),
            ('id', 'child::cim:requirement/cim:requirement/cim:lateralBoundaryCondition/cim:id'),
            ('id', 'child::cim:requirement/cim:requirement/cim:outputRequirement/cim:id'),
            ('id', 'child::cim:requirement/cim:requirement/cim:spatioTemporalConstraint/cim:id'),
            # Metafor CMIP5 questionnaire - simulation bundle
            ('name', 'child::cim:requirement/cim:requirement/cim:boundaryCondition/cim:name'),
            ('name', 'child::cim:requirement/cim:requirement/cim:initialCondition/cim:name'),
            ('name', 'child::cim:requirement/cim:requirement/cim:lateralBoundaryCondition/cim:name'),
            ('name', 'child::cim:requirement/cim:requirement/cim:outputRequirement/cim:name'),
            ('name', 'child::cim:requirement/cim:requirement/cim:spatioTemporalConstraint/cim:name'),
        ]
    }


def boundary_condition():
    """A boundary condition is a numerical requirement which looks like a variable imposed on the model evolution (i.e. it might - or might not - evolve with time, but is seen by the model at various times during its evolution) as opposed to an initial condition (at model time zero).

    """
    return {
        'type': 'class',
        'base': 'activity.numerical_requirement',
        'is_abstract': False,
        'constraints': [
            ('requirement_type', 'constant', 'boundaryCondition'),
        ]
    }


def initial_condition():
    """An initial condition is a numerical requirement on a model prognostic variable value at time zero.

    """
    return {
        'type': 'class',
        'base': 'activity.numerical_requirement',
        'is_abstract': False,
        'constraints': [
            ('requirement_type', 'constant', 'initialCondition'),
        ]
    }


def spatio_temporal_constraint():
    """Contains a set of relationship types specific to a simulation document that can be used to describe its genealogy.

    """
    return {
        'type': 'class',
        'base': 'activity.numerical_requirement',
        'is_abstract': False,
        'properties': [
            ('date_range', 'shared.date_range', '0.1'),
            ('spatial_resolution', 'activity.resolution_type', '0.1'),
        ],
        'constraints': [
            ('requirement_type', 'constant', 'spatioTemporalConstraint'),
        ],
        'decodings': [
            ('date_range', 'child::cim:requiredDuration/cim:closedDateRange', 'shared.closed_date_range'),
            ('date_range', 'child::cim:requiredDuration/cim:openDateRange', 'shared.open_date_range'),
            ('spatial_resolution', 'child::cim:spatialResolution'),
        ]
    }


def output_requirement():
    """Contains a set of relationship types specific to a simulation document that can be used to describe its genealogy.

    """
    return {
        'type': 'class',
        'base': 'activity.numerical_requirement',
        'is_abstract': False,
        'constraints': [
            ('requirement_type', 'constant', 'outputRequirement'),
        ]
    }


def physical_modification():
    """Physical modification is the implementation of a boundary condition numerical requirement that is achieved within the model code rather than from some external source file. It  might include, for example,  a specific rate constant within a chemical reaction, or coefficient value(s) in a parameterisation.  For example, one might require a numerical experiment where specific chemical reactions were turned off - e.g. no heterogeneous chemistry.

    """
    return {
        'type': 'class',
        'base': 'activity.conformance',
        'is_abstract': False
    }


def simulation():
    """A simulation is the implementation of a numerical experiment.  A simulation can be made up of "child" simulations aggregated together to form a simulation composite.  The parent simulation can be made up of whole or partial child simulations, the simulation attributes need to be able to capture this.

    """
    return {
        'type': 'class',
        'base': 'activity.numerical_activity',
        'is_abstract': True,
        'properties': [
            ('authors', 'str', '0.1'),
            ('calendar', 'shared.calendar', '1.1'),
            ('control_simulation', 'activity.simulation', '0.1'),
            ('conformances', 'activity.conformance', '0.N'),
            ('deployments', 'software.deployment', '0.N'),
            ('inputs', 'software.coupling', '0.N'),
            ('outputs', 'data.data_object', '0.N'),
            ('restarts', 'data.data_object', '0.N'),
            ('simulation_id', 'str', '0.1'),
            ('spinup_date_range', 'shared.closed_date_range', '0.1'),
            ('spinup_simulation', 'activity.simulation', '0.1'),
        ],
        'doc_strings': {
            'authors': 'List of associated authors.',
            'control_simulation': 'Points to a simulation being used as the basis (control) run.  Note that only "derived" simulations can describe something as being control; a simulation should not know if it is being used itself as the control of some other run.',
            'inputs': 'Implemented as a mapping from a source to target; can be a forcing file, a boundary condition, etc.',
            'spinup_date_range': 'The date range that a simulation is engaged in spinup.',
            'spinup_simulation': 'The (external) simulation used during "spinup."  Note that this element can be used in conjuntion with spinupDateRange.  If a simulation has the latter but not the former, then one can assume that the simulation is performing its own spinup.',
        },
        'decodings': [
            ('calendar', 'child::cim:calendar/cim:daily-360', 'shared.daily_360'),
            ('calendar', 'child::cim:calendar/cim:perpetualPeriod', 'shared.perpetual_period'),
            ('calendar', 'child::cim:calendar/cim:realCalendar', 'shared.real_calendar'),
            ('authors', 'child::cim:authorsList/cim:list'),
            ('conformances', 'child::cim:conformance/cim:conformance', 'activity.conformance'),
            ('conformances', 'child::cim:conformance/cim:physicalModification', 'activity.physical_modification'),
            ('control_simulation', 'child::cim:controlSimulation/cim:controlSimulation'),
            ('control_simulation', 'child::cim:controlSimulation/cim:reference', 'shared.doc_reference'),
            ('deployments', 'child::cim:deployment'),
            ('inputs', 'child::cim:input'),
            ('simulation_id', 'child::cim:simulationID'),
            ('spinup_date_range', 'child::cim:dateRange/cim:closedDateRange'),
        ]
    }


def simulation_composite():
    """A SimulationComposite is an aggregation of Simulations. With the aggreation connector between Simulation and SimulationComposite(SC) the SC can be made up of both SimulationRuns and SCs. The SimulationComposite is the new name for the concept of SimulationCollection: A simulation can be made up of "child" simulations aggregated together to form a "simulation composite".  The "parent" simulation can be made up of whole or partial child simulations and the SimulationComposite attributes need to be able to capture this.

    """
    return {
        'type': 'class',
        'base': 'activity.simulation',
        'is_abstract': False,
        'is_document': True,
        'properties': [
            ('child', 'activity.simulation', '0.N'),
            ('date_range', 'shared.date_range', '1.1'),
            ('rank', 'int', '1.1'),
        ],
        'doc_strings': {
            'rank': 'Position of a simulation in the SimulationComposite timeline. eg:  Is this the first (rank = 1) or second (rank = 2) simulation'
        },
        'decodings': [
            ('meta', 'self::cim:simulationRun'),
            ('child', 'child::cim:child', 'activity.simulation_run'),
            ('child', 'child::cim:child', 'activity.simulation_composite'),
            ('date_range', 'child::cim:dateRange/cim:closedDateRange', 'shared.closed_date_range'),
            ('date_range', 'child::cim:dateRange/cim:openDateRange', 'shared.open_date_range'),
            ('rank', 'child::cim:rank'),
        ]
    }


def simulation_relationship():
    """Contains a set of relationship types specific to a simulation document that can be used to describe its genealogy.

    """
    return {
        'type': 'class',
        'base': 'shared.relationship',
        'is_abstract': False,
        'properties': [
            ('target', 'activity.simulation_relationship_target', '1.1'),
            ('type', 'activity.simulation_relationship_type', '1.1'),
        ],
    }


def simulation_relationship_target():
    """Creates and returns instance of simulation_relationship_target class.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('simulation', 'shared.doc_reference', '0.1'),
            ('target', 'activity.simulation_type', '0.1')
        ],
    }


def simulation_run():
    """A SimulationRun is, as the name implies, one single model run. A SimulationRun is a Simulation. There is a one to one association between SimulationRun and (a top-level) SoftwarePackage::ModelComponent.

    """
    return {
        'type': 'class',
        'base': 'activity.simulation',
        'is_abstract': False,
        'is_document': True,
        'properties': [
            ('date_range', 'shared.date_range', '1.1'),
            ('model', 'software.model_component', '0.1')
        ],
        'decodings': [
            ('meta', 'self::cim:simulationRun'),
            ('date_range', 'child::cim:dateRange/cim:closedDateRange', 'shared.closed_date_range'),
            ('date_range', 'child::cim:dateRange/cim:openDateRange', 'shared.open_date_range'),
            ('model', 'child::cim:model/cim:modelComponent'),
            ('model', 'child::cim:model/cim:reference', 'shared.doc_reference')
        ]
    }



# -*- coding: utf-8 -*-

"""
.. module:: cim.v1.decoder_for_quality_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: A set of cim 1 decoders.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
from decoder_xml_utils import set_attributes
from decoder_for_shared_package import *
import typeset



def decode_cim_quality(xml, nsmap):
    """Decodes an instance of the following type: cim quality.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.quality.CimQuality

    """
    decodings = [
        ('meta', False, decode_doc_meta_info, 'self::cim:cIM_Quality'),
        ('reports', True, decode_report, 'child::cim:report'),
    ]

    return set_attributes(typeset.quality.CimQuality(), xml, nsmap, decodings)


def decode_evaluation(xml, nsmap):
    """Decodes an instance of the following type: evaluation.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.quality.Evaluation

    """
    decodings = [
        ('date', False, 'datetime.datetime', 'child::gmd:result/gmd:DQ_ConformanceResult/gmd:specification/gmd:CI_Citation/gmd:date/gmd:CI_Date/gmd:date/gco:Date'),
        ('description', False, 'unicode', 'gmd:evaluationMethodDescription/gco:CharacterString'),
        ('did_pass', False, 'bool', 'child::gmd:result/gmd:DQ_ConformanceResult/gmd:pass/gco:Boolean'),
        ('explanation', False, 'unicode', 'child::gmd:result/gmd:DQ_ConformanceResult/gmd:explanation/gco:CharacterString'),
        ('specification', False, 'unicode', 'child::gmd:result/gmd:DQ_ConformanceResult/gmd:specification/@xlink:title'),
        ('specification_hyperlink', False, 'unicode', 'child::gmd:result/gmd:DQ_ConformanceResult/gmd:specification/@xlink:href'),
        ('title', False, 'unicode', 'child::gmd:result/gmd:DQ_ConformanceResult/gmd:specification/gmd:CI_Citation/gmd:title/gco:CharacterString'),
        ('type', False, 'unicode', 'child::gmd:result/@xlink:title'),
        ('type_hyperlink', False, 'unicode', 'child::gmd:result/@xlink:href'),
    ]

    return set_attributes(typeset.quality.Evaluation(), xml, nsmap, decodings)


def decode_measure(xml, nsmap):
    """Decodes an instance of the following type: measure.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.quality.Measure

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:measureDescription'),
        ('description', False, 'unicode', 'parent::cim:report/child::gmd:measureDescription/gco:CharacterString'),
        ('identification', False, 'unicode', 'child::cim:measureIdentification/gmd:code/gco:CharacterString'),
        ('name', False, 'unicode', 'child::cim:nameOfMeasure'),
        ('name', False, 'unicode', 'parent::cim:report/child::gmd:nameOfMeasure/gco:CharacterString'),
    ]

    return set_attributes(typeset.quality.Measure(), xml, nsmap, decodings)


def decode_report(xml, nsmap):
    """Decodes an instance of the following type: report.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.quality.Report

    """
    decodings = [
        ('date', False, 'datetime.datetime', 'child::gmd:dateTime/gco:DateTime'),
        ('evaluation', False, decode_evaluation, 'self::cim:report'),
        ('evaluator', False, decode_responsible_party, 'child::cim:evaluator'),
        ('measure', False, decode_measure, 'self::cim:report/cim:measure'),
    ]

    return set_attributes(typeset.quality.Report(), xml, nsmap, decodings)





# -*- coding: utf-8 -*-

"""
.. module:: quality_enums.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Set of CIM v1 quality package enum definitions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""



def cim_feature_type():
    """Creates and returns instance of cim_feature_type enum.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ('file', None),
            ('diagnostic', None),
        ],
    }


def cim_result_type():
    """Creates and returns instance of cim_result_type enum.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ('plot', None),
            ('document', None),
            ('logfile', None),
        ],
    }


def cim_scope_code_type():
    """This would cover quality issues with the CIM itself.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ('dataset', None),
            ('software', None),
            ('service', None),
            ('model', None),
            ('modelComponent', None),
            ('simulation', None),
            ('experiment', None),
            ('numericalRequirement', None),
            ('ensemble', None),
            ('file', None),
        ],
    }


def quality_issue_type():
    """Creates and returns instance of quality_issue_type enum.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ('metadata', None),
            ('data_format', None),
            ('data_content', None),
            ('data_indexing', None),
            ('science', None),
        ],
    }


def quality_severity_type():
    """Creates and returns instance of quality_severity_type enum.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ('cosmetic', None),
            ('minor', None),
            ('major', None),
        ],
    }


def quality_status_type():
    """Creates and returns instance of quality_status_type enum.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ('reported', None),
            ('confirmed', None),
            ('partially_resolved', None),
            ('resolved', None),
        ],
    }



from pyesdoc.mp.ontologies.generators.qconfig import utils
from pyesdoc.mp.ontologies.generators.qconfig.qconfig_generator import QConfigGenerator

# Expose utility functions.
UTILS = utils

# Set of supported custom generators or tornado templates.
GENERATORS = {
    QConfigGenerator
}



# -*- coding: utf-8 -*-

"""
.. module:: default.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Default document extender.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc import constants
from pyesdoc.ontologies import cim



def _set_full_id(ctx):
    """Sets full document identifier.

    """
    ctx.ext.full_id = "{}-{}".format(ctx.meta.id, ctx.meta.version)


def _set_type_info(ctx):
    """Sets document type information.

    """
    ctx.meta.type = ctx.ext.type = ctx.doc.__class__.type_key
    ctx.ext.css_class = ctx.ext.type.lower().replace(".", "-")
    ctx.meta.type_display_name = cim.constants.DISPLAY_NAMES.get(type(ctx.doc), ctx.meta.type.split(".")[-1])
    ctx.meta.type_sort_key = cim.constants.SORT_KEYS.get(type(ctx.doc), u"ZZ")


def _set_institute(ctx):
    """Sets document institute field.

    """
    if ctx.meta.institute:
        ctx.meta.institute = ctx.meta.institute.upper()
    else:
        ctx.meta.institute = constants.DEFAULT_INSTITUTE


def _set_project(ctx):
    """Sets document project field.

    """
    if ctx.meta.project:
        ctx.meta.project = ctx.meta.project.lower()
    else:
        ctx.meta.institute = constants.DEFAULT_PROJECT


def _set_summary_fields(ctx):
    """Sets document summary fields.

    """
    fields = ()
    if hasattr(ctx.doc, "short_name"):
        fields = fields + (ctx.doc.short_name, )
    elif hasattr(ctx.doc, "canonical_name"):
        fields = fields + (ctx.doc.canonical_name, )
    if hasattr(ctx.doc, "long_name"):
        fields = fields + (ctx.doc.long_name, )
    ctx.ext.summary_fields = fields


def _set_display_name(ctx):
    """Sets document display name.

    """
    ctx.ext.display_name = None
    if hasattr(ctx.doc, "short_name") and ctx.doc.short_name:
        ctx.ext.display_name = ctx.doc.short_name
    elif hasattr(ctx.doc, "name") and ctx.doc.name:
        ctx.ext.display_name = ctx.doc.name
    elif hasattr(ctx.doc, "title") and ctx.doc.title:
        ctx.ext.display_name = ctx.doc.title


def _set_description(ctx):
    """Sets document description.

    """
    ctx.ext.description = None
    if hasattr(ctx.doc, "description") and ctx.doc.description:
        ctx.ext.description = ctx.doc.description[:1023]
    elif hasattr(ctx.doc, "long_name") and ctx.doc.long_name:
        ctx.ext.description = ctx.doc.long_name[:1023]


def _set_viewer_url(ctx):
    """Sets document viewer URL.

    """
    ctx.ext.viewer_url = constants.VIEWER_URL_BY_ID.format(ctx.meta.project,
                                                           ctx.meta.id,
                                                           ctx.meta.version)

def _set_full_display_name(ctx):
    """Sets document full display name.

    """
    # Escape if already assigned.
    if ctx.ext.full_display_name:
        return

    name = u""
    if ctx.meta.project:
        name += ctx.meta.project.upper()
        name += u" "
    name += ctx.meta.type_display_name
    name += u" : "
    if ctx.meta.institute and ctx.meta.institute != u"--":
        name += ctx.meta.institute.upper()
        name += u" - "
    if ctx.ext.display_name:
        name += ctx.ext.display_name
        name += u" "

    ctx.ext.full_display_name = name.strip()


def _set_sort_key(ctx):
    """Sets document full display name.

    """
    ctx.meta.sort_key = u"{0}{1}".format(
        ctx.meta.type_sort_key, ctx.ext.display_name or unicode())
    ctx.meta.sort_key = ctx.meta.sort_key.upper()


# Set of pre-extenders.
PRE_EXTENDERS = (
    _set_full_id,
    _set_type_info,
    _set_institute,
    _set_project,
    _set_display_name,
    _set_description,
    _set_summary_fields,
    _set_viewer_url,
    )


# Set of post-extenders.
POST_EXTENDERS = (
    _set_full_display_name,
    _set_sort_key
    )




# -*- coding: utf-8 -*-
"""
.. module:: handlers_dkrz_qc.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Exposes DKRZ QC document processing handlers.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""

def _set_institute(doc):
    """Sets institute code.

    """
    # Escape if no external id has been defined.
    if len(doc.meta.external_ids) == 0:
        return

    # Derive DRS.
    drs = doc.meta.external_ids[0].value
    drs = drs.split('.')[2:]

    # Derive institute from DRS.
    if len(drs) > 0:
        doc.meta.institute = drs[0].upper()


# Set of document parsers.
PARSERS = (
    _set_institute,
    )



# -*- coding: utf-8 -*-

"""
.. module:: grids_enums.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Set of CIM v1 grids package enum definitions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""


def arc_type_enum():
    """Creates and returns instance of arc_type_enum enum.

    """
    return {
        'type' : 'enum',
        'is_open' : False,
        'members' : [
            ('geodesic', None),
            ('great_circle', None),
            ('small_circle', None),
            ('complex', None)
        ],
    }


def discretization_enum():
    """Creates and returns instance of discretization_enum enum.

    """
    return {
        'type' : 'enum',
        'is_open' : False,
        'members' : [
            ('logically_rectangular', None),
            ('structured_triangular', None),
            ('unstructured_triangular', None),
            ('pixel-based_catchment', None),
            ('unstructured_polygonal', None),
            ('spherical_harmonics', None),
            ('other', None)
        ],
    }


def feature_type_enum():
    """Creates and returns instance of feature_type_enum enum.

    """
    return {
        'type' : 'enum',
        'is_open' : False,
        'members' : [
            ('point', None),
            ('edge', None)
        ],
    }


def geometry_type_enum():
    """Creates and returns instance of geometry_type_enum enum.

    """
    return {
        'type' : 'enum',
        'is_open' : False,
        'members' : [
            ('ellipsoid', None),
            ('plane', None),
            ('sphere', None)
        ],
    }


def grid_node_position_enum():
    """Creates and returns instance of grid_node_position_enum enum.

    """
    return {
        'type' : 'enum',
        'is_open' : False,
        'members' : [
            ('centre', None),
            ('plane', None),
            ('sphere', None)
        ],
    }


def grid_type_enum():
    """Creates and returns instance of grid_type_enum enum.

    """
    return {
        'type' : 'enum',
        'is_open' : False,
        'members' : [
            ('cubed_sphere', None),
            ('displaced_pole', None),
            ('icosahedral_geodesic', None),
            ('reduced_gaussian', None),
            ('regular_lat_lon', None),
            ('spectral_gaussian', None),
            ('tripolar', None),
            ('yin_yang', None),
            ('composite', None),
            ('other', None)
        ],
    }


def horizontal_cs_enum():
    """Creates and returns instance of horizontal_cs_enum enum.

    """
    return {
        'type' : 'enum',
        'is_open' : False,
        'members' : [
            ('cartesian', None),
            ('ellipsoidal', None),
            ('polar', None),
            ('spherical', None)
        ],
    }


def refinement_type_enum():
    """Creates and returns instance of refinement_type_enum enum.

    """
    return {
        'type' : 'enum',
        'is_open' : False,
        'members' : [
            ('none', 'Tile boundaries have no refinement when the grid lines meeting at the tile boundary are continuous.'),
            ('integer', 'The refinement is integer when grid lines from the coarser grid are continuous on the finer grid, but not vice versa.'),
            ('rational','The refinement is rational when the adjacent or overlapping grid tiles have grid line counts that are coprime (i.e. no common factor other than 1).')
        ],
    }


def vertical_cs_enum():
    """Creates and returns instance of vertical_cs_enum enum.

    """
    return {
        'type' : 'enum',
        'is_open' : False,
        'members' : [
            ('mass-based', None),
            ('space-based', None)
        ],
    }



# -*- coding: utf-8 -*-

"""
.. module:: test_type_cim_v1_quality_cim_quality.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Tests a cim.v1.CimQuality instance.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import arrow

import pyesdoc.ontologies.cim as cim
import test_utils as tu



# Test type.
DOC_TYPE = cim.v1.CimQuality

# Test type display name.
DOC_TYPE_DISPLAY_NAME = "Dataset QC"

# Test document type.
DOC_TYPE_KEY = DOC_TYPE.type_key

# Test representation file.
DOC_FILE = 'cim.1.quality.CimQuality.xml-metafor-cim-v1'

# Test document uid.
DOC_UID = 'c1debb66-2737-11e2-bc06-0010185b3f28'

# Test document version.
DOC_VERSION = '2'

# Test document creation date.
DOC_DATE = arrow.get('2012-11-05T09:51:25').datetime

# Test document project.
DOC_PROJECT = "CMIP5"

# Test document project.
DOC_INSTITUTE = "MOHC"

# Test document author.
DOC_AUTHOR = "Metafor Questionnaire"


def assert_extension_info(ext):
    """Asserts a document's extension information.

    :param object ext: Document extension information.

    """
    tu.assert_str(ext.display_name, "CMIP5 Dataset Quality Control Report")
    tu.assert_str(ext.description, "CMIP5 Quality Control", True)
    tu.assert_str(ext.full_display_name, " CMIP5 Dataset Quality Control Report")
    tu.assert_int(ext.summary_fields, 1)
    tu.assert_str(ext.summary_fields[0], "cmip5.output1.", True)


def _assert_doc_core(doc, is_update):
    """Assert core information."""
    tu.assert_iter(doc.reports, 1)
    r = doc.reports[0]
    tu.assert_date(r.date, '2011-05-01 12:00:00')
    tu.assert_object(r.evaluator, cim.v1.ResponsibleParty)
    tu.assert_str(r.evaluator.individual_name, 'stockhause@dkrz.de')
    tu.assert_str(r.evaluator.role, 'pointofContact')


def _assert_doc_report_evaluation(doc, is_update):
    """Assert report evaluation information."""
    r = doc.reports[0]
    tu.assert_object(r.evaluation, cim.v1.Evaluation)
    e = r.evaluation
    tu.assert_date(e.date, '2012-11-05')
    tu.assert_str(e.description, 'evaluationMethodType=indirect', True)
    tu.assert_str(e.explanation, 'The quality results for this ESG', True)
    tu.assert_str(e.specification, 'The CMIP5/AR5 Model Data Quality Control')
    tu.assert_str(e.specification_hyperlink, 'http://cmip5qc.wdc-climate.de')
    tu.assert_str(e.type, 'QC Level 2 Results')
    tu.assert_str(e.type_hyperlink, 'http://cera-www.dkrz.de/WDCC/CMIP5/QCResult.jsp?experiment=cmip5/output1/NCC/NorESM1-ME/rcp45')


def _assert_doc_report_measure(doc, is_update):
    """Assert report measure information."""
    r = doc.reports[0]
    tu.assert_object(r.measure, cim.v1.Measure)
    m = r.measure
    tu.assert_str(m.description, 'WDCC Conformance', True)
    tu.assert_str(m.identification, 'cmip5-qc-2d')
    tu.assert_str(m.name, 'CMIP5 Quality Control Data Level 2')


def assert_doc(doc, is_update=False):
    """Asserts a document.

    :param object doc: Document being tested.
    :param bool is_update: Flag indicating whether document has been updated.

    """
    for assertor in (
        _assert_doc_core,
        _assert_doc_report_evaluation,
        _assert_doc_report_measure,
        ):
        assertor(doc, is_update)


def update_doc(doc):
    """Update a document prior to republishing.

    :param object doc: Document being republished.

    """
    pass


def assert_doc_updates(doc):
    """Asserts a document after being updated.

    :param object doc: Document being tested.

    """
    pass


# -*- coding: utf-8 -*-
"""
.. module:: pyesdoc.ontologies.utils.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encpasulates utility functions related to registered ontologies.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc import constants
from pyesdoc.ontologies import type_info



def get_type_key(name, version, package, typeof):
    """Returns type key.

    :param str name: Ontology name, e.g. cim.
    :param str version: Ontology version, e.g. 1.
    :param str package: Ontology package, e.g. activity.
    :param str typeof: Ontology type, e.g. Experiment.

    :returns: A type key.
    :rtype: str

    """
    return "{0}.{1}.{2}.{3}".format(name, version, package, typeof).lower()


def get_types(name=None, version=None):
    """Returns set of supported types.

    :param str name: Ontology name, e.g. cim.
    :param str version: Ontology version, e.g. 1.

    :returns: A tuple of supported types.
    :rtype: tuple

    """
    result = ()

    for o in type_info.ONTOLOGIES:
        if name is None or name == o.NAME:
            for v in o.VERSIONS:
                if version is None or version == v.ID:
                    result += v.type_info.TYPES

    return result


def list_types(name=None, version=None):
    """Returns list of supported type keys and types.

    :param str name: Ontology name, e.g. cim.
    :param str version: Ontology version, e.g. 1.

    :returns: A tuple of supported typekeys and types.
    :rtype: tuple

    """
    types = get_types(name, version)
    types = [tuple(t.type_key.split('.')) for t in types]

    return tuple(types)


def get_doc_type_keyset():
    """Returns set of supported document type keys.

    :returns: Set of supported document type keys.
    :rtype: set

    """
    return type_info.KEYS.values()


def get_type(name, version, package, typeof):
    """Returns a type if supported.

    :param str name: Ontology name, e.g. cim.
    :param str version: Ontology version, e.g. 1.
    :param str package: Ontology package, e.g. activity.
    :param str typeof: Ontology type, e.g. Experiment.

    :returns: A type (if found).
    :rtype: class or None

    """
    return get_type_from_key(get_type_key(name, version, package, typeof))


def get_type_key(doc_type):
    """Returns a document type key.

    :param class doc_type: A document type for which meta-information is to be returned.

    :returns: A type key (if found).
    :rtype: str or None

    """
    return type_info.KEYS.get(doc_type, None)


def get_type_from_key(key):
    """Returns a type if supported.

    :param str key: Type key, e.g. cim.1.software.ModelComponent.

    :returns: A type (if found).
    :rtype: class or None

    """
    type_key = str(key).lower()
    for t in type_info.TYPES:
        if type_info.KEYS[t].lower() == type_key:
            return t


def is_supported(name, version, package=None, typeof=None):
    """Returns a flag indicating whether ontology/type is supported.

    :param str name: Ontology name, e.g. cim.
    :param str version: Ontology version, e.g. 1.
    :param str package: Ontology package, e.g. activity.
    :param str typeof: Ontology type, e.g. Experiment.

    :returns: A flag indicating whether ontology is supported.
    :rtype: bool

    """
    if (package is not None and typeof is None) or \
       (package is None and typeof is not None):
       raise ValueError("The ontology package and type are unspecified.")

    if package is None and typeof is None:
        return len(get_types(name, version)) > 0
    else:
        return get_type(name, version, package, typeof) is not None


def get_decoder_info(doc_type):
    """Returns meta-information associated with a type used for decoding purposes.

    :param class doc_type: A document type for which meta-information is to be returned.

    :returns: Type meta-information.
    :rtype: tuple

    """
    result = {name: typeof for name, _, typeof in get_constraints(doc_type, constants.CONSTRAINT_TYPE_TYPEOF)}
    for name, _, cardinality in get_constraints(doc_type, constants.CONSTRAINT_TYPE_CARDINALITY):
        result[name] = (name, result[name], bool(cardinality.split(".")[1] == 'N'))

    return result.values()


def get_constraint(constraints, constraint_type):
    """Returns first constraint within collection of the matching type.

    :param list constraints: A collection of constraints.
    :param str constraint_type: Type of constraint.

    :returns: Constrain meta-information.
    :rtype: tuple

    """
    for typeof, value in constraints:
        if typeof == constraint_type:
            return value


def get_constraints(doc_type, type_filter=None):
    """Returns constraints associated with a type.

    :param class doc_type: A document type for which meta-information is to be returned.
    :param str type_filtre: A constraint type filter.

    :returns: Type meta-information.
    :rtype: tuple

    """
    if type_filter:
        return [ct for ct in type_info.CONSTRAINTS[doc_type] if ct[1] == type_filter]

    return type_info.CONSTRAINTS[doc_type]


def get_property_constraint(container, name, typeof):
    """Returns property constraints associated with a type instance.

    :param class container: A document instance for which meta-information is to be returned.
    :param str name: Property name.
    :param str typeof: Type of property constraint to be returned.

    :returns: Constraint meta-information.
    :rtype: tuple

    """
    return get_constraint(type_info.CONSTRAINTS[type(container), name], typeof)




# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.schema.utils
   :platform: Unix, Windows
   :synopsis: Ontology schema utility functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.mp.ontologies.schemas import cim



def get_schema(name, version):
    """Returns a supported ontology schema module.

    :param str name: Schema name.
    :param str version: Schema version.

    :returns: An ontology schema.
    :rtype: module

    """
    if name == 'cim':
        if version == '1':
            return cim.v1
        if version == '2':
            return cim.v2

    raise KeyError("Ontology schema not found: {} :: v{}".format(name, version))


def parse_type(typeof):
    """Parses a type declaration.

    """
    if typeof.startswith('linked_to'):
        parts = [i.strip() for i in typeof[10:-1].split(",")]
        try:
            return parts[0], parts[1]
        except IndexError:
            return parts[0], None
    else:
        return typeof, None



"""

A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.
"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotte Pascoe'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Charlotte Pascoe, Robert Pincus'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Characteristics of the atmosphere radiation process'

# --------------------------------------------------------------------
# PROCESS: top level properties
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': "Radiative agents in the atmosphere",
    'properties': [
        ('aerosols', 'ENUM:aerosol_types', '1.N',
            'Aerosols whose radiative effect is taken into account in the atmosphere model'),
        ('greenhouse_gases', 'ENUM:ghg_types', '1.N',
            'Greenhouse gases whose radiative effect is taken into account in the atmosphere model'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: shortwave_radiation
# --------------------------------------------------------------------
DETAILS['shortwave_radiation'] = {
    'description': 'Properties of the shortwave radiation scheme',
    'properties': [
        ('name', 'str', '0.1',
            'Commonly used name for the shortwave radiation scheme'),
        ('spectral_integration', 'ENUM:spectral_integration', '1.1',
            'Shortwave radiation scheme spectral integration'),
        ('transport_calculation', 'ENUM:transport_calculation', '1.N',
            'Shortwave radiation transport calculation methods'),
        ('spectral_intervals', 'int', '1.1',
            'Shortwave radiation scheme number of spectral intervals'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: sw_cloud_ice
# --------------------------------------------------------------------
DETAILS['sw_cloud_ice'] = {
    'description': 'Shortwave radiative properties of ice crystals in clouds',
    'properties': [
        ('general_interactions', 'ENUM:radiative_interactions', '1.N',
            'General shortwave radiative interactions with cloud ice crystals'),
        ('physical_representation', 'ENUM:cloud_ice_physical_representation', '1.N',
            'Physical representation of cloud ice crystals in the shortwave radiation scheme'),
        ('optical_methods', 'ENUM:optical_methods', '1.N',
            'Optical methods applicable to cloud ice crystals in the shortwave radiation scheme'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: sw_cloud_liquid
# --------------------------------------------------------------------
DETAILS['sw_cloud_liquid'] = {
    'description': 'Shortwave radiative properties of liquid droplets in clouds',
    'properties': [
        ('general_interactions', 'ENUM:radiative_interactions', '1.N',
            'General shortwave radiative interactions with cloud liquid droplets'),
        ('physical_representation', 'ENUM:cloud_liquid_physical_representation', '1.N',
            'Physical representation of cloud liquid droplets in the shortwave radiation scheme'),
        ('optical_methods', 'ENUM:optical_methods', '1.N',
            'Optical methods applicable to cloud liquid droplets in the shortwave radiation scheme'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: sw_aerosols
# --------------------------------------------------------------------
DETAILS['sw_aerosols'] = {
    'description': 'Shortwave radiative properties of aerosols',
    'properties': [
        ('general_interactions', 'ENUM:radiative_interactions', '1.N',
            'General shortwave radiative interactions with aerosols'),
        ('physical_representation', 'ENUM:aerosol_physical_representation', '1.N',
            'Physical representation of aerosols in the shortwave radiation scheme'),
        ('optical_methods', 'ENUM:optical_methods', '1.N',
            'Optical methods applicable to aerosols in the shortwave radiation scheme'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: sw_gases
# --------------------------------------------------------------------
DETAILS['sw_gases'] = {
    'description': 'Shortwave radiative properties of gases',
    'properties': [
        ('general_interactions', 'ENUM:radiative_interactions', '1.N',
            'General shortwave radiative interactions with gases'),
        ('optical_methods', 'ENUM:optical_methods', '1.N',
            'Optical methods applicable to gases in the shortwave radiation scheme'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: longwave_radiation
# --------------------------------------------------------------------
DETAILS['longwave_radiation'] = {
    'description': 'Properties of the longwave radiation scheme',
    'properties': [
        ('name', 'str', '0.1',
            'Commonly used name for the longwave radiation scheme.'),
        ('spectral_integration', 'ENUM:spectral_integration', '1.1',
            'Longwave radiation scheme spectral integration'),
        ('transport_calculation', 'ENUM:transport_calculation', '1.N',
            'Longwave radiation transport calculation methods'),
        ('spectral_intervals', 'int', '1.1',
            'Longwave radiation scheme number of spectral intervals'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: lw_cloud_ice
# --------------------------------------------------------------------
DETAILS['lw_cloud_ice'] = {
    'description': 'Longwave radiative properties of ice crystals in clouds',
    'properties': [
        ('general_interactions', 'ENUM:radiative_interactions', '1.N',
            'General longwave radiative interactions with cloud ice crystals'),
        ('physical_reprenstation', 'ENUM:cloud_ice_physical_representation', '1.N',
            'Physical representation of cloud ice crystals in the longwave radiation scheme'),
        ('optical_methods', 'ENUM:optical_methods', '1.N',
            'Optical methods applicable to cloud ice crystals in the longwave radiation scheme'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: lw_cloud_liquid
# --------------------------------------------------------------------
DETAILS['lw_cloud_liquid'] = {
    'description': 'Longwave radiative properties of liquid droplets in clouds',
    'properties': [
        ('general_interactions', 'ENUM:radiative_interactions', '1.N',
            'General longwave radiative interactions with cloud liquid droplets'),
        ('physical_representation', 'ENUM:cloud_liquid_physical_representation', '1.N',
            'Physical representation of cloud liquid droplets in the longwave radiation scheme'),
        ('optical_methods', 'ENUM:optical_methods', '1.N',
            'Optical methods applicable to cloud liquid droplets in the longwave radiation scheme'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: lw_aerosols
# --------------------------------------------------------------------
DETAILS['lw_aerosols'] = {
    'description': 'Longwave radiative properties of aerosols',
    'properties': [
        ('general_interactions', 'ENUM:radiative_interactions', '1.N',
            'General longwave radiative interactions with aerosols'),
        ('physical_representation', 'ENUM:aerosol_physical_representation', '1.N',
            'Physical representation of aerosols in the longwave radiation scheme'),
        ('optical_methods', 'ENUM:optical_methods', '1.N',
            'Optical methods applicable to aerosols in the longwave radiation scheme'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: lw_gases
# --------------------------------------------------------------------
DETAILS['lw_gases'] = {
    'description': 'Longwave radiative properties of gases',
    'properties': [
        ('general_interactions', 'ENUM:radiative_interactions', '1.N',
            'General longwave radiative interactions with gases'),
        ('optical_methods', 'ENUM:optical_methods', '1.N',
            'Optical methods applicable to gases in the longwave radiation scheme'),
    ]
}

# --------------------------------------------------------------------
# PROCESS: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['aerosol_types'] = {
    'description': 'Aerosols whose radiative effect is taken into account in the atmospheric model.',
    'is_open': True,
    'members': [
        ('sulphate', None),
        ('nitrate', None),
        ('sea salt', None),
        ('dust', None),
        ('ice', None),
        ('organic', None),
        ('BC (black carbon / soot)', None),
        ('SOA (secondary organic aerosols)', None),
        ('POM (particulate organic matter)', None),
        ('polar stratospheric ice', None),
        ('NAT (nitric acid trihydrate)', None),
        ('NAD (nitric acid dihydrate)', None),
        ('STS (supercooled ternary solution aerosol particle)', None),
        ]
    }

ENUMERATIONS['ghg_types'] = {
    'description': 'Greenhouse gases whose radiative effect is taken into account in the atmospheric model',
    'is_open': True,
    'members': [
        ('H2O', None),
        ('CO2', None),
        ('CH4', None),
        ('N2O', None),
        ('O3', None),
        ('CFCs', None),
        ('HFCs', None),
        ('PFCs', None),
        ('SF6', None),
        ('NF3', None),
        ]
    }

ENUMERATIONS['spectral_integration'] = {
    'description': 'Spectral integration of the radiation scheme',
    'is_open': True,
    'members': [
        ('wide-band model', None),
        ('correlated-k', None),
        ('exponential sum fitting', None),
        ]
    }

ENUMERATIONS['transport_calculation'] = {
    'description': 'Radiation transport calculation methods',
    'is_open': True,
    'members': [
        ('two-stream', None),
        ('layer interaction', None),
        ('bulk', 'highly parameterised methods that use bulk expressions'),
        ('adaptive', 'exploits spatial and temporal correlations in optical characteristics'),
        ('multi-stream', None)
        ]
    }

ENUMERATIONS['radiative_interactions'] = {
    'description': 'General radiative interactions',
    'is_open': True,
    'members': [
        ('scattering', None),
        ('emission/absorption', None),
    ]
}

ENUMERATIONS['cloud_ice_physical_representation'] = {
    'description': 'Physical representation of ice crystals in clouds',
    'is_open': True,
    'members': [
        ('bi-modal size distribution',
            'small mode diameters: a few tens of microns, large mode diameters: typically hundreds of microns'),
        ('ensemble of ice crystals', 'complex shapes represented with an ensemble of symmetric shapes'),
        ('mean projected area',
            'randomly oriented irregular ice crystals present a greater mean projected area than spheres'),
        ('ice water path', 'Integrated ice water path through the cloud kg m-2',),
        ('crystal asymmetry', None),
        ('crystal aspect ratio', None),
        ('effective crystal radius', None),
        ]
    }

ENUMERATIONS['cloud_liquid_physical_representation'] = {
    'description': 'Physical_radiative representation of liquid droplets in clouds',
    'is_open': True,
    'members': [
        ('cloud droplet number concentration', 'CDNC'),
        ('effective cloud droplet radii', None),
        ('droplet size distribution', None),
        ('liquid water path', 'Integrated liquid water path through the cloud kg m-2',),
        ]
    }

ENUMERATIONS['aerosol_physical_representation'] = {
    'description': 'Physical radiative representation of aerosols',
    'is_open': True,
    'members': [
        ('number concentration', None),
        ('effective radii', None),
        ('size distribution', None),
        ('asymmetry', None),
        ('aspect ratio', None),
        ('mixing state', 'For shortwave radiative interaction'),
        ]
    }

ENUMERATIONS['optical_methods'] = {
    'description': 'Optical methods used by radiation scheme',
    'is_open': True,
    'members': [
        ('T-matrix', 'For non-spherical particles'),
        ('geometric optics', 'For non-spherical particles'),
        ('finite difference time domain (FDTD)', 'For non-spherical particles'),
        ('Mie theory', 'For spherical particles'),
        ('anomalous diffraction approximation', None),
        ]
    }

ENUMERATIONS['cloud_ice_radiation_processes'] = {
    'description': 'Optical radiative processes for ice crystals in clouds',
    'is_open': True,
    'members': [
        ('emissivity', None),
        ('absorption', None),
        ('backward scattering', None),
        ('side scattering', None),
        ]
    }

ENUMERATIONS['cloud_liquid_radiation_processes'] = {
    'description': 'Optical radiative processes for liquid droplets in clouds',
    'is_open': True,
    'members': [
        ('droplet scattering', None),
        ('droplet absorption', None),
        ('broadband reflectivity', 'albedo'),
        ('broadband transmissivity', None),
        ('broadband absorbtivity', None),
        ]
    }

ENUMERATIONS['shortwave_scheme_type'] = {
    'description': 'Type of scheme used for shortwave radiation parameterisation',
    'is_open': True,
    'members': [
        ('wide-band model', None),
        ('bulk-scheme', 'highly parameterised methods that use bulk expressions'),
        ('two-stream', None),
        ('two-stream (delta-Eddington)', 'approximation for solar radiation calculations'),
        ]
    }

ENUMERATIONS['single_scattering_properties_methods'] = {
    'description': 'Methods for calculating single scattering properties of atmospheric constituents',
    'is_open': True,
    'members': [
        ('T-Matrix', 'an exact method'),
        ('geometrical optics', 'for particles that are much larger than the wavelength of light'),
        ('finite difference time domain', 'FDTD'),
        ('anomalous diffraction approximation', 'ADA'),
        ('k-distribution', None),
        ('band model', None),
        ('exponential sum fitting', None),
    ]
}






"""
.. module:: factory.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Takes specialization modules and returns instances of wrapper classes.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from utils_constants import *
import utils_model as model



# Map of specializations by id.
CACHE = dict()



def get_specialization(modules):
    """Returns a specialization wrapper.

    :param modules: 4 member tuple of python modules: root, grid, key-properties, processes.

    :returns: A specialization wrapper.
    :rtype: tuple

    """
    # Unpack modules.
    mod_root, mod_grid, mod_keyprops, mod_processes = modules

    # Set root type key.
    type_key = TYPE_KEY_REALM if mod_root.__name__ != "model" else TYPE_KEY_MODEL

    # Create topic tree.
    result = _create_topic(None, mod_root, type_key)
    if mod_grid is not None:
        _create_topic(result, mod_grid, TYPE_KEY_GRID)
    _create_topic(result, mod_keyprops, TYPE_KEY_KEYPROPS)
    for i in mod_processes:
        _create_topic(result, i, TYPE_KEY_PROCESS)

    return result


def _create_topic(parent, spec, type_key, key=None):
    """Creates & returns a topic specialization wrapper.

    """
    if spec is None:
        return None

    # Instantiate.
    topic = model.TopicSpecialization(spec, type_key)
    if isinstance(spec, dict):
        _set_topic_from_dict(parent, topic, key)
    else:
        _set_topic_from_module(parent, topic)

    # Update hierachy.
    if parent:
        topic.parent = parent
        parent.sub_topics.append(topic)

    # Cache (used in downstream tooling chain).
    CACHE[topic.id] = topic

    return topic


def _set_topic_from_module(parent, topic):
    """Set topic specialization attributes from a module.

    """
    topic.authors = topic.spec.AUTHORS
    topic.contact = topic.spec.CONTACT
    topic.qc_status = topic.spec.QC_STATUS
    topic.description = topic.spec.DESCRIPTION
    if parent:
        topic.change_history = parent.change_history
        topic.contributors = parent.contributors
        topic.name = "_".join(topic.spec.__name__.split(".")[-1].split("_")[1:])
        topic.id = "{}.{}".format(parent.id, topic.name)
    else:
        topic.change_history = topic.spec.CHANGE_HISTORY
        topic.contributors = topic.spec.CONTRIBUTORS
        topic.name = topic.spec.__name__
        topic.id = "cmip6.{}".format(topic.name)

    # Assign properties / property sets.
    if hasattr(topic.spec, "DETAILS") and hasattr(topic.spec, "ENUMERATIONS"):
        for key, obj in topic.spec.DETAILS.items():
            # ... toplevel properties
            if key == "toplevel":
                _set_property_collection(topic, obj, topic.spec.ENUMERATIONS)

            # ... toplevel property sets
            elif key.startswith("toplevel"):
                _set_property_set(topic, key, obj, topic.spec.ENUMERATIONS)

            # ... sub-topic properties
            elif len(key.split(":")) == 1:
                _create_topic(topic, obj, TYPE_KEY_SUBPROCESS, key)
                _set_property_collection(topic.sub_topics[-1], obj, topic.spec.ENUMERATIONS)

            # ... sub-topic property sets
            elif len(key.split(":")) == 2:
                for st in topic.sub_topics:
                    if st.name == key.split(":")[0]:
                        _set_property_set(st, key, obj, topic.spec.ENUMERATIONS)


def _set_topic_from_dict(parent, topic, name):
    """Set topic specialization attributes from a dictionary.

    """
    topic.authors = parent.authors
    topic.contact = parent.contact
    topic.change_history = parent.change_history
    topic.contributors = parent.contributors
    topic.description = topic.spec['description']
    topic.id = "{}.{}".format(parent.id, name)
    topic.name = name
    topic.qc_status = parent.qc_status


def _set_property_set(owner, key, obj, enumerations):
    """Set attributes of a property-set attributes from a dictionary.

    """
    ps = model.PropertySetSpecialization()
    ps.description = obj['description']
    ps.id = "{}.{}".format(owner.id, key.split(":")[-1])
    ps.key = key
    ps.name = key.split(":")[-1]
    ps.owner = owner
    _set_property_collection(ps, obj, enumerations)

    owner.property_sets.append(ps)

    # Cache.
    CACHE[ps.id] = ps


def _set_property_collection(owner, obj, enumerations):
    """Set a collection of topic properties from a dictionary.

    """
    for name, typeof, cardinality, description in obj.get('properties', []):
        p = model.PropertySpecialization()
        p.cardinality = cardinality
        p.description = description
        p.enum = _create_enum(p, typeof, enumerations) if typeof.startswith("ENUM:") else None
        p.id = "{}.{}".format(owner.id, name)
        p.key = name
        p.name = name
        p.owner = owner
        p.typeof = typeof

        owner.properties.append(p)

        # Cache.
        CACHE[p.id] = p


def _create_enum(detail, typeof, enumerations):
    """Creates & returns an enumeration specialzation wrapper.

    """
    key = typeof.split(":")[-1]
    obj = enumerations[key]

    e = model.EnumSpecialization()
    e.description = obj['description']
    e.detail = detail
    e.id = "{}.{}".format(detail.id, key)
    e.is_open = obj['is_open']
    e.label = key
    e.name = key
    e.id = key
    e.choices = [_create_enum_choice(e, i[0], i[1]) for i in obj.get('members', [])]

    return e


def _create_enum_choice(enum, value, description):
    """Creates & returns an enumeration choice specialzation wrapper.

    """
    ec = model.EnumChoiceSpecialization()
    ec.description = description
    ec.enum = enum
    ec.id = "{}.{}".format(enum.id, value)
    ec.value = value

    return ec




# -*- coding: utf-8 -*-

"""
.. module:: test_validation.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyesdoc document validation tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
# Module imports.
import pyesdoc
import test_utils as tu
import test_types as tt
from pyesdoc.utils import runtime as rt



def _set_doc_errors_01(doc):
    """Valid document - i.e. error count = 0."""
    return 0


def _set_doc_errors_02(doc):
    """Invalid document - meta attribute is missing."""
    doc.meta = None

    return 1


def _test(mod):
    """Performs standard document validation tests."""
    # Set test document/error factories.
    error_setters = (
        _set_doc_errors_01,
        _set_doc_errors_02
        )
    if hasattr(mod, "get_document_error_setters"):
        error_setters += mod.get_document_error_setters()

    # Perform validation test.
    for error_setter in error_setters:
        # ... set valid test document.
        doc = tu.get_doc(mod)

        # ... set expected error count.
        expected_error_count = error_setter(doc)

        # ... validate document.
        errors = pyesdoc.validate(doc)

        # ... print errors if mismatch between expected & actual.
        # if len(errors) == expected_error_count:
        #     for error in errors:
        #         rt.log(error, level=rt.LOG_LEVEL_WARNING)

        # ... assert errors.
        tu.assert_iter(errors, expected_error_count, length_compare=tu.COMPARE_LTE)


def test():
    """Performs validation tests over the set of test documents."""
    for mod in tt.MODULES:
        tu.init(_test, 'validation', mod)
        yield _test, mod


# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.generators.generator_utils
   :platform: Unix, Windows
   :synopsis: Set of common generator utility functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import datetime
import os
import pwd

import tornado



# Templates folder.
_TEMPLATE_FOLDER = os.path.dirname(__file__)

# Standard 4 character python indent.
_INDENT = '    '

# Standard line return.
_LINE_RETURN = '\n'

# Set of loaded templates.
_loaded_templates = dict()

# Set of loaded tornado templates.
_loaded_tornado_templates = dict()


def convert_to_camel_case(name, separator='_'):
    """Converts passed name to camel case.

    :param str name: A name as specified in ontology specification.
    :param str separator: Separator to use in order to split name into constituent parts.

    """
    r = ''
    if name is not None:
        s = name.split(separator)
        for s in s:
            if (len(s) > 0):
                r += s[0].upper()
                if (len(s) > 1):
                    r += s[1:]
    return r


def convert_to_pascal_case(name, separator='_'):
    """Converts passed name to camel case.

    :param str name: A name as specified in ontology specification.
    :param str separator: Separator to use in order to split name into constituent parts.

    """
    r = ''
    s = convert_to_camel_case(name, separator)
    if (len(s) > 0):
        r += s[0].lower()
        if (len(s) > 1):
            r += s[1:]
    return r


def emit(iterable, func=None, sort=True, joiner=""):
    """Returns code (sorted).

    """
    if len(iterable) == 0:
        return ""

    if func:
        code = [func(i) for i in iterable]
    else:
        code = iterable
    code = [l for l in code if l is not None]

    return joiner.join(sorted(code)) if sort else joiner.join(code)


def _load_template(language, filename):
    """Returns code template.

    :param str language: Generator language.
    :param str filename: Name of template file.

    """
    path = _TEMPLATE_FOLDER + "/{0}/templates/{1}".format(language, filename)

    if path not in _loaded_templates:
        tmpl = open(path)
        _loaded_templates[path] = tmpl.read()
        tmpl.close()

    return _loaded_templates[path]


def load_templates(language, filenames):
    """Returns a dictionary of loaded code templates.

    :param str language: Generator language.
    :param str filenames: Set of template file names.

    """
    templates = {}
    for filename in filenames:
        templates[filename] = _load_template(language, filename)

    return templates


def load_tornado_template(language, fname):
    """Returns tornado code template.

    :param str language: Generator language.
    :param str fname: Name of template file.

    """
    dir_ = _TEMPLATE_FOLDER + "/{0}/templates".format(language)

    fpath = os.path.join(dir_, fname)
    if fpath not in _loaded_tornado_templates:
        loader = tornado.template.Loader(dir_)
        _loaded_tornado_templates[fpath] = loader.load(fpath)

    return _loaded_tornado_templates[fpath]


def load_tornado_templates(language, fnames):
    """Returns a dictionary of loaded tornado code templates.

    :param str language: Generator language.
    :param str fnames: Set of template file names.

    """
    return {fname: load_tornado_template(language, fname) for fname in fnames}


def get_username():
    """Returns name of current user.

    """
    try:
        return pwd.getpwuid(os.getuid()).pw_name
    except ImportError:
        return os.environ.get("USERNAME")


def emit_indent(count=1):
    """Emits code corresponding to a code indentation.

    :param int count: Number of indentations to emit.

    """
    return reduce(lambda x, y: x + _INDENT, range(count), '')


def emit_tab():
    """Emits code corresponding to a code tab.

    """
    return emit_indent(4)


def emit_line_return(count=1):
    """Emits code corresponding to a code line return.

    :param int count: Number of line returns to emit.

    """
    return reduce(lambda x, y: x + _LINE_RETURN, range(count), '')


def create_directory(dir):
    """Generates a directory into which code will be generated.

    :param str dir: Target code generation directory.

    """
    try:
        os.makedirs(dir)
    except:
        pass


def write_file(code, dir, file):
    """Writes code to a file.

    :param str code: Code to be written to a file.
    :param str dir: Directory into which code is to be generated.
    :param str file: Name of code file being written.

    """
    # Create directory.
    create_directory(dir)

    # Update code.
    code = code.replace('{file-name}', file)

    # Write file.
    file = open(dir + "/" + file, 'w')
    file.writelines(code)
    file.close()


def format_code(ctx, code):
    """Formats code prior to being written to file system.

    :param pyesdoc.mp.ontologies.generators.generator.GeneratorContext ctx: Generation context information.
    :param str code: Code to be injected with standard params.

    """
    # Ontology related params.
    code = code.replace('{ontology-name}', ctx.ontology.op_name)
    code = code.replace('{ontology-version}', ctx.ontology.op_version)
    code = code.replace('{ontology-version-packagename}', ctx.ontology.op_version.replace('.', '_'))

    # Misceallaneous params.
    code = code.replace('{datetime-year}', str(datetime.datetime.now().year))
    code = code.replace('{user-name}', get_username())

    return code



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.drq.definition.xml_parser.py
   :copyright: Copyright "February 27, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates parsing over the data request definition XML file.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.drq import constants
from pyesdoc.drq import utils



class XMLParser(object):
    """Event driven data request definition XML file parser.

    """
    def execute(self):
        """Execute the parse.

        """
        self.on_parse_begin()

        # Load definition xml file.
        fpath = utils.get_fpath(constants.DEFINITION_FNAME, "definition")
        xml = utils.load_xml(fpath)

        # Parse.
        for elem in xml.findall('./table'):
            self.on_parse_table_element(elem)
            for sub_elem in elem.findall('./rowAttribute'):
                self.on_parse_table_attribute_element(elem, sub_elem)

        self.on_parse_complete()


    def on_parse_begin(self):
        """On parse begin event handler.

        """
        pass


    def on_parse_table_element(self, elem):
        """On table element parse event handler.

        """
        pass


    def on_parse_table_attribute_element(self, table_elem, elem):
        """On table attribute row element parse event handler.

        """
        pass


    def on_parse_complete(self):
        """On parse complete event handler.

        """
        pass



"""

A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.
"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotte Pascoe'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Charlotte Pascoe'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Characteristics of the cloud simulator'

# --------------------------------------------------------------------
# SUB-PROCESSES
# --------------------------------------------------------------------
DETAILS['isscp_attributes'] = {
    'description': 'ISSCP Characteristics',
    'properties': [
        ('top_height', 'ENUM:isscp_top_height', '1.N',
            'Cloud simulator ISSCP top height'),
        ('top_height_direction', 'ENUM:isscp_top_height_direction', '1.1',
            'Cloud simulator ISSCP top height direction'),
        ]
    }

DETAILS['cosp_attributes'] = {
    'description': 'CFMIP Observational Simulator Package attributes',
    'properties': [
        ('run_configuration', 'ENUM:cosp_run_configuration', '1.1',
            'Cloud simulator COSP run configuration'),
        ('number_of_grid_points', 'int', '1.1',
            'Cloud simulator COSP number of grid points'),
        ('number_of_columns', 'int', '1.1',
            'Cloud simulator COSP number of cloumns'),
        ('number_of_levels', 'int', '1.1',
             'Cloud simulator COSP number of levels'),
        ]
    }

DETAILS['radar_inputs'] = {
    'description': 'Characteristics of the cloud radar simulator',
    'properties': [
        ('frequency', 'float', '1.1',
         'Cloud simulator radar frequency (Hz)'),
        ('type', 'ENUM:inputs_radar_type', '1.1',
         'Cloud simulator radar type'),
        ('gas_absorption', 'bool', '1.1',
         'Cloud simulator radar uses gas absorption'),
        ('effective_radius', 'bool', '1.1',
         'Cloud simulator radar uses effective radius'),
        ]
    }

DETAILS['lidar_inputs'] = {
    'description': 'Characteristics of the cloud lidar simulator',
    'properties': [
        ('ice_types', 'ENUM:inputs_lidar_ice_type', '1.1',
         'Cloud simulator lidar ice type'),
        ('overlap', 'ENUM:inputs_lidar_overlap', '1.N',
         'Cloud simulator lidar overlap'),
        ]
    }

# --------------------------------------------------------------------
# PROCESS: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['isscp_top_height'] = {
    'description': 'Cloud top height management',
    'is_open': True,
    'members': [
        ('no adjustment', None),
        ('IR brightness', None),
        ('visible optical depth', None),
        ]
    }

ENUMERATIONS['isscp_top_height_direction'] = {
    'description': 'Direction for finding the radiance determined cloud-top pressure. '
                   'Atmosphere pressure level with interpolated temperature equal to '
                   'the radiance determined cloud-top pressure.',
    'is_open': True,
    'members': [
        ('lowest altitude level', None),
        ('highest altitude level', None),
        ]
    }

ENUMERATIONS['cosp_run_configuration'] = {
    'description': 'Method used to run the CFMIP Observational Simulator Package',
    'is_open': True,
    'members': [
        ('Inline', None),
        ('Offline', None),
        ]
    }

ENUMERATIONS['inputs_radar_type'] = {
    'description': 'Type of radar',
    'is_open': True,
    'members': [
        ('surface', None),
        ('space borne', None),
        ]
    }

ENUMERATIONS['inputs_lidar_ice_type'] = {
    'description': 'Ice particle shape in lidar calculations',
    'is_open': True,
    'members': [
        ('ice spheres', None),
        ('ice non-spherical', None),
        ]
    }

ENUMERATIONS['inputs_lidar_overlap'] = {
    'description': 'lidar overlap type',
    'is_open': True,
    'members': [
        ('max', None),
        ('random', None),
        ]
    }



# -*- coding: utf-8 -*-

"""
.. module:: designing_classes.py
   :synopsis: Set of CIM v2 ontology type definitions.

"""


def axis_member():
    """PLACEHOLDER for the real axis_member.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
        ]
    }


def domain_requirements():
    """Properties of the domain which needs to be simulated, extent and/or resolution.

    """
    return {
        'type': 'class',
        'base': 'designing.numerical_requirement',
        'is_abstract': False,
        'properties': [
            ('required_extent', 'science.topic', '0.1',
                "Constraint on extent of domain to be simulated."),
            ('required_resolution', 'science.topic', '0.1',
                "Constraint on resolution required in simulated domain.")
        ],
        'constraints': [
            ('cardinality', 'additional_requirements', '0.0')
        ]
    }


def ensemble_requirement():
    """Defines an experiment ensemble.

    """
    return {
        'type': 'class',
        'base': 'designing.numerical_requirement',
        'is_abstract': False,
        'properties': [
            ('ensemble_member', 'linked_to(designing.numerical_requirement)', '0.N',
                "Constraint on each ensemble member."),
            ('ensemble_type', 'designing.ensemble_types', '1.1',
                "Type of ensemble."),
            ('minimum_size', 'int', '1.1',
                "Minimum number of members.")
        ],
        'constraints': [
            ('cardinality', 'additional_requirements', '0.0')
        ]
    }


def ensemble_types():
    """Defines the various axes along which one can set up an ensemble, whether
     as an experiment designer, or in designing a 'response' ensemble around an
     experiment.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ("Perturbed Physics", "Members differ in some aspects of their physics."),
            ("Initialisation Method", "Members differ in how they are initialised."),
            ("Realization", "Members initialised to sample possible initial conditions."),
            ("Start Date", "Members initialised at different starting dates."),
            ("Forced", "Members have differing forcing data."),
            ("Resolution", "Members are/should-be run at different resolutions."),
            ("Driven", "Members are/should-be driven by different models.")
        ]
    }


def forcing_constraint():
    """Identifies a model forcing constraint.

    """
    return {
        'type': 'class',
        'base': 'designing.numerical_requirement',
        'is_abstract': False,
        'properties': [
            ('additional_constraint', 'str', '0.1',
                "Additional information, e.g. hold constant from 2100-01-01."),
            ('category', 'str', '0.1',
                "Category to which this belongs (from a CV, e.g. GASES)."),
            ('code', 'str', '0.1',
                "Programme wide code from a controlled vocabulary (e.g. N2O)."),
            ('data_link', 'data.dataset', '0.1',
                "A data record used by the forcing "),
            ('forcing_type', 'designing.forcing_types', '1.1',
                "Type of integration."),
            ('group', 'str', '0.1',
                "Sub-Category (e.g. GHG)."),
            ('origin', 'shared.citation', '0.1',
                "Pointer to origin, e.g. CMIP6 RCP database.")
        ],
        'constraints': [
            ('cardinality', 'additional_requirements', '0.0')
        ]
    }


def forcing_types():
    """Defines the possible set of forcing types for a forcing constraint.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ("historical", "Best estimates of actual state (included synthesized)"),
            ("idealised", "Simplified and/or exemplar, e.g. 1%C02"),
            ("scenario", "Intended to represent a possible future, e.g. RCP4.5"),
            ("driven", "Driven from another simulation")
        ]
    }


def initialisation_requirement():
    """A requirement on how a particular simulation should be initialised.

    """
    return {
        'type': 'class',
        'base': 'designing.numerical_requirement',
        'is_abstract': False,
        'properties': [
            ('branch_time_in_initialisation_source', 'time.date_time', '0.1',
                "If appropriate,  the time in the initialisation_source (whether observed or simulated)."),
            ('initialise_from_data', 'data.dataset', '0.1',
                "Initialisation should use this primary dataset."),
            ('initialise_from_experiment', 'designing.numerical_experiment', '0.1',
                "This experiment should be initialised from the output of this experiment.")
        ],
        'constraints': [
            ('mutually_exclusive', ('initialise_from_experiment', 'initialise_from_data'), '1.1')
        ]
    }


def multi_ensemble():
    """In the case of multiple ensemble axes, defines how they
    are set up and ordered.

    """
    return {
        'type': 'class',
        'base': 'designing.numerical_requirement',
        'is_abstract': False,
        'properties': [
            ('ensemble_axis', 'linked_to(designing.ensemble_requirement)', '1.N',
                "List of orthogonal ensembles.")
        ],
        'constraints': [
            ('cardinality', 'additional_requirements', '0.0')
        ]
    }


def numerical_experiment():
    """Defines a numerical experiment.

    """
    return {
        'type': 'class',
        'base': 'activity.activity',
        'is_abstract': False,
        'properties': [
            ('tier', 'int', '0.1',
                "Relative importance of experiment within a MIP."),
            ('related_experiments', 'designing.numerical_experiment', '0.N',
                "Other experiments which have defined relationships to this one."),
            ('governing_mips', 'linked_to(designing.project)', '0.N',
                "MIP(s) overseeing experimental design protocol."),
            ('related_mips', 'linked_to(designing.project)', '0.N',
                "MIP's that require this experiment."),
            ('required_period', 'linked_to(designing.temporal_constraint)', '1.1',
                "Constraint on start date and duration."),
            ('requirements', 'linked_to(designing.numerical_requirement)', '0.N',
                "Additional requirements that conformant simulations need to satisfy.")
        ],
        'constraints': [
            ('cardinality', 'duration', '0.0'),
            ('cardinality', 'rationale', '1.1')
        ]
    }


def experimental_relationships():
    """Defines the canonical set of experimental relationships.

    """
    return {
        'type': 'enum',
        'is_open': True,
        'members': [
            ('is_constrained_by',
                "The experiment that provides constraint(s) for the target experiment (e.g SST forcing)."),
            ('is_constrainer_of',
                "The set of experiments constrained by the experiment."),
            ('is_perturbation_from',
                "The experiment that provides a control for the target experiment."),
            ('is_control_for',
                "The set of experiments that use the experiment as a control."),
            ('is_initialized_by',
                "The experiment providing initialization for the experiment."),
            ('is_initializer_of',
                "The set of experiments initialized by the experiment."),
            ('is_sibling_of',
                "Part of a family (e.g. experiments where solar forcing is either increased or reduced)."),
        ]
    }


def numerical_requirement():
    """A numerical requirement associated with a numerical experiment.

    """
    return {
        'type': 'class',
        'base': 'activity.activity',
        'is_abstract': False,
        'alternatives': [
            'designing.temporal_constraint',
            'designing.forcing_constraint',
            'designing.ensemble_requirement',
            'designing.multi_ensemble',
            'designing.start_date_ensemble',
            'designing.output_requirement',
            'designing.domain_requirements',
            'designing.initialisation_requirement'
        ],
        'properties': [
            ('additional_requirements', 'linked_to(designing.numerical_requirement)', '0.N',
                "Additional detail for this requirement."),
            ('is_conformance_requested', 'bool', '1.1',
                "Indicator as to whether ensemble documentation should include conformance information for this requirement."),
            ('is_conformance_info_required', 'bool', '1.1',
                "Indicator as to whether an institute is expected to provide informationas to how it confirmed."),
            ('scope', 'designing.numerical_requirement_scope', '0.1',
                "Scope allows us to categorise a requirement in terms of how widely it is shared.")
        ],
        'constraints': [
            ('cardinality', 'duration', '0.0')
        ]
    }


def numerical_requirement_scope():
    """The scope to which a numerical requirement may or may not apply.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ("mip-era", "MIP era wide e.g. 'concentration of pre-industrial CO2' & 'Impose AMIP SSTs'"),
            ("mip-group", "Shared across a group of MIPs e.g. aerosol forcing in GeoMIP and AerChemMIP."),
            ("mip", "Shared within a single MIP e.g. spin-up protocol for land surface experiments in LUMIP."),
            ("experiment", "Applies to a single experiment e.g. CFMIP 'zonally uniform SST plus 4K'")
        ]
    }


def output_requirement():
    """Provides details of what output is required and when from an experiment.

    """
    return {
        'type': 'class',
        'base': 'designing.numerical_requirement',
        'is_abstract': False,
        'properties': [
            ('formal_data_request', 'shared.online_resource', '0.1',
                "If available, link to a 'cmip' style online request.")
        ],
        'constraints': [
            ('cardinality', 'additional_requirements', '0.0')
        ]
    }


def project():
    """Describes a scientific project.

    """
    return {
        'type': 'class',
        'base': 'activity.activity',
        'is_abstract': False,
        'properties': [
            ('homepage', 'str', '0.1',
                "Project homepage."),
            ('objectives', 'str', '0.N',
                "Project objectives."),
            ('previous_projects', 'linked_to(designing.project)', '0.N',
                "Previous projects with similar aims."),
            ('required_experiments', 'linked_to(designing.numerical_experiment)', '0.N',
                "Experiments required to deliver project."),
            ('governed_experiments', 'linked_to(designing.numerical_experiment)', '0.N',
                "Experiments governed by this project."),
            ('sub_projects', 'linked_to(designing.project)', '0.N',
                "Activities within the project with their own name and aim(s).")
        ],
        'constraints': [
            ('cardinality', 'description', '1.1')
        ]
    }


def simulation_plan():
    """Describes a simulation that needs to be run.

    """
    return {
        'type': 'class',
        'base': 'activity.activity',
        'is_abstract': False,
        'properties': [
            ('expected_model', 'linked_to(science.model)', '1.1',
                "The model used to run the simulation."),
            ('expected_performance_sypd', 'float', '0.1',
                "Expected performance in simulated years per real day."),
            ('expected_platform', 'linked_to(platform.machine)', '0.1',
                "The machine on which the simulation will be run."),
            ('will_support_experiments', 'linked_to(designing.numerical_experiment)', '1.N',
                "An experiment with which the planned simulation will be associated.")
        ],
        'constraints': [
            ('cardinality', 'duration', '1.1')
        ]
    }


def start_date_ensemble():
    """Defines an experiment ensemble with multiple start dates.

    """
    return {
        'type': 'class',
        'base': 'designing.numerical_requirement',
        'is_abstract': False,
        'properties': [
            ('ensemble_members', 'time.datetime_set', '1.1',
                "Description of date or time set of start dates.")
        ],
        'constraints': [
            ('cardinality', 'additional_requirements', '0.0')
        ]
    }


def temporal_constraint():
    """A spatio-temporal constraint on a numerical experiment.

    """
    return {
        'type': 'class',
        'base': 'designing.numerical_requirement',
        'is_abstract': False,
        'properties': [
            ('required_calendar', 'time.calendar', '0.1',
                "Required calendar (e.g. for paleo simulations)."),
            ('required_duration', 'time.time_period', '0.1',
                "Constraint on time or length of simulation."),
            ('start_date', 'time.date_time', '0.1',
                "Required start date."),
            ('start_flexibility', 'time.time_period', '0.1',
                "Amount of time before required start date that it is permissible to begin integration.")
        ],
        'constraints': [
            ('cardinality', 'additional_requirements', '0.0')
        ]
    }



# -*- coding: utf-8 -*-

"""
.. module:: activity_classes.py
   :synopsis: Set of CIM v2 ontology type definitions.

"""


def activity():
    """Base class for activities.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': True,
        'is_document': True,
        'properties': [
            ('alternative_names', 'str', '0.N',
                "List of names by which the activity is also known."),
            ('canonical_name', 'str', '0.1',
                "Community defined identifier or name."),
            ('citations', 'shared.citation', '0.N',
                "Set of pertinent citations."),
            ('description', 'str', '0.1',
                "Description of what is to be done (or was done)."),
            ('duration', 'time.time_period', '0.1',
                "Time the activity was (or will be) active."),
            ('internal_name', 'str', '0.1',
                "A name used for internal purposes."),
            ('keywords', 'str', '0.1',
                "User defined keywords."),
            ('long_name', 'str', '0.1',
                "Longer version of activity name."),
            ('name', 'str', '1.1',
                "Short name or abbreviation."),
            ('responsible_parties', 'linked_to(shared.responsibility)', '0.N',
                "People or organisations responsible for activity."),
            ('previously_known_as', 'str', '0.N',
                "List of names by which the activity was formerly known."),
            ('rationale', 'str', '0.1',
                "Explanation of why this activity was carried out and/or what it was intended to achieve.")
        ]
    }


def axis_member():
    """Description of a given ensemble member. It will normally be related to a specific
    ensemble requirement. Note that start dates can be extracted directly from the simulations
    and do not need to be recorded with an axis member description.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('description', 'str', '1.1',
                "Description of the member (or name of parameter varied)."),
            ('extra_detail', 'str', '0.1',
                "If necessary: further information about ensemble member conformance."),
            ('index', 'int', '1.1',
                "The ensemble member index."),
            ('value', 'float', '0.1',
                "If parameter varied, value thereof for this member.")
        ]
    }


def conformance():
    """A specific conformance. Describes how a particular numerical requirement has been implemented.
    Will normally be linked from an ensemble descriptor.

    """
    return {
        'type': 'class',
        'base': 'activity.activity',
        'is_abstract': False,
        'properties': [
            ('conformance_achieved', 'activity.conformance_type', '1.1',
                "Summary of conformance status."),
            ('datasets', 'data.input_dataset', '0.N',
                "The datasets (including any modifications made to them) used for conforming to the target requirement."),
            ('target_requirement', 'designing.numerical_requirement', '1.1',
                "URI of the target numerical requirement."),
            ('models', 'science.model', '1.N',
                 'The models to which this conformance applies.'),
        ],
        'constraints': [
            ('cardinality', 'rationale', '0.0'),
            ('cardinality', 'duration', '0.0'),
            ('cardinality', 'canonical_name', '0.0'),
            ('cardinality', 'keywords', '0.0'),
            ('cardinality', 'description', '1.1')
        ]
    }


def conformance_type():
    """Standardised set of conformance responses.

    """
    return {
        'type': 'enum',
        'is_open': True,
        'members': [
            ("Conformed", "Simulation (or ensemble) conformed to requirement"),
            ("Partially Conformed", "Simulation (or ensemble) partially conformed to requirement - details in description"),
            ("Not Conformed", "Simulation (or ensemble) was unable to conform to requirement"),
            ("Not Applicable", "Requirement is not relevant")
        ]
    }


def ensemble():
    """Generic ensemble definition.
    Holds the definition of how the various ensemble members have been configured.
    If ensemble axes are not present, then this is either a single member ensemble,
    or part of an uber ensemble.

    """
    return {
        'type': 'class',
        'base': 'activity.activity',
        'is_abstract': False,
        'properties': [
            ('common_conformances', 'activity.conformance', '0.N',
                "Conformance documents for requirements common across ensemble."),
            ('common_performance', 'platform.performance', '0.1',
                "Representative model performance across ensemble."),
            ('documentation', 'linked_to(shared.online_resource)', '0.N',
                "Links to web-pages and other ensemble specific documentation (including workflow descriptions)."),
            ('ensemble_axes', 'activity.ensemble_axis', '0.N',
                "Set of axes for the ensemble."),
            ('members', 'activity.ensemble_member', '1.N',
                "The set of ensemble members."),
            ('uber_ensembles', 'activity.uber_ensemble', '0.N',
                "Link to one or more over-arching ensembles that might includes this one."),
            ('experiments', 'designing.numerical_experiment', '1.N',
                "Experiments with which the ensemble is associated (may differ from constituent simulations).")
        ],
        'constraints': [
            ('cardinality', 'rationale', '0.0'),
            ('cardinality', 'canonical_name', '0.0'),
            ('cardinality', 'keywords', '0.0'),
            ('cardinality', 'duration', '0.0')
        ]
    }


def ensemble_axis():
    """Defines the meaning of r/i/p indices in an ensemble.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'pstr': ('{}', ('axis',)),
        'properties': [
            ('extra_detail', 'str', '1.1',
                "Any extra detail required to describe how this ensemble axis was delivered."),
            ('member', 'activity.axis_member', '1.N',
                "Individual member descriptions along axis."),
            ('short_identifier', 'str', '1.1',
                "e.g. 'r' or 'i' or 'p' to conform with simulation ensemble variant identifiers."),
            ('target_requirement', 'linked_to(designing.numerical_requirement)', '1.1',
                "URI of the target numerical requirement.")
        ]
    }


def ensemble_member():
    """An ensemble may be a complicated interplay of axes, for example, r/i/p, not all of which
    are populated, so we need a list of the actual simulations and how they map onto the ensemble
    axes.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('errata', 'linked_to(shared.online_resource)', '0.1',
                "Link to errata associated with this simulation."),
            ('had_performance', 'linked_to(platform.performance)', '0.1',
                "Performance of the simulation."),
            ('ran_on', 'linked_to(platform.machine)', '0.1',
                "The machine on which the simulation was run."),
            ('simulation', 'linked_to(data.simulation)', '1.1',
                "Actual simulation description for an ensemble member. The variant id is in the simuluation document.")
        ]
    }


def parent_simulation():
    """Defines the relationship between a simulation and its parent.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('branch_method', 'str', '0.1',
                "Description of how the simulation was branched from a parent simualtion, e.g. 'standard', 'none provided'."),
            ('branch_time_in_child', 'time.date_time', '0.1',
                "The time at which the present simulation started in the child calendar."),
            ('branch_time_in_parent', 'time.date_time', '0.1',
                "The time in parent simulation calendar at which this simulation was branched."),
            ('parent', 'linked_to(data.simulation)', '1.1',
                "The parent simulation of this child.")
        ]
    }


def uber_ensemble():
    """An ensemble made up of other ensembles. Often used where parts of an ensemble were run by
    different institutes. Could also be used when a new experiment is designed which can use
    ensemble members from previous experiments and/or projects.

    """
    return {
        'type': 'class',
        'base': 'activity.ensemble',
        'is_abstract': False,
        'properties': [
            ('child_ensembles', 'linked_to(activity.ensemble)', '1.N',
                "Ensemble which are aggregated into this one.")
        ],
        'constraints': [
            ('cardinality', 'ensemble_axes', '1.N'),
            ('cardinality', 'common_conformances', '0.0'),
            ('cardinality', 'members', '0.0')
        ]
    }



# -*- coding: utf-8 -*-
"""
.. module:: config.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Configuration utility functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.utils import config



def get_project_sources():
    """Returns iterable of unique project/source pairs.

    """
    config.init()

    result = []
    for project in get_projects():
        for source in (f.source for f in project.feeds):
            if (project.name, source) not in result:
                result.append((project.name, source))

    return tuple(result)


def get_directory():
    """Returns archive directory as defined withn pyesdoc.conf.

    """
    config.init()

    return config.data.archive.location


def get_projects():
    """Returns set of archive projects.

    """
    config.init()

    return config.data.archive.projects


def get_sources():
    """Returns set of archive sources.

    """
    config.init()

    return config.data.archive.sources




# -*- coding: utf-8 -*-

"""
.. module:: data_data_object.py
   :platform: Unix, Windows
   :synopsis: Extends a cim.v1.data.data_object document.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""



def get_extenders():
    """Returns set of extension functions.

    """
    return (
        _set_display_name,
        _set_summary_fields
        )


def _set_display_name(ctx):
    """Sets document display name.

    """
    ctx.ext.display_name = ctx.doc.acronym


def _set_summary_fields(ctx):
    """Sets document summary fields.

    """
    ctx.ext.summary_fields += (
        ctx.doc.acronym,
        ctx.doc.description
    )



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.drq.definition.table.py
   :copyright: Copyright "February 27, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Wraps a table defined within dreq2Defn.xml.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.drq import constants
from pyesdoc.drq import utils



# Map of table xml attribute names to value covertors.
_CONVERTORS = {
    'is_label_unique': lambda v: False if v == "No" else True,
    'level': int,
    'max_occurs': int
}



class Table(object):
    """Wraps a table defined within dreq2Defn.xml.

    """
    def __init__(self, elem):
        """Instance constructor.

        :param xml.etree.Element elem: XML element declared within data request definition.

        """
        utils.init_from_xml(self, elem, elem.keys(), _CONVERTORS)
        self.attributes = []
        self.label_drq = self.label
        self.label = utils.get_label(self.label)


    def __repr__(self):
        """Instance representation.

        """
        return self.label


    def __len__(self):
        """Returns number of attributes in managed collection.

        """
        return len(self.attributes)


    def __contains__(self, target):
        """Returns flag indicating whether an attribute of the matching name exists or not.

        """
        return (target in self.attributes) or (self[target] is not None)


    def __iter__(self):
        """Instance iterator initializer.

        """
        return iter(sorted(self.attributes, key=lambda i: i._sort_key))


    def __getitem__(self, key):
        """Returns a child table attribute.

        """
        if isinstance(key, int):
            return self.attributes[key]
        key = str(key).strip().lower()
        for attribute in self.attributes:
            if attribute.label.lower() == key or \
               attribute.label_drq.lower() == key:
                return attribute



# -*- coding: utf-8 -*-

"""
.. module:: activity_simulation_run.py
   :platform: Unix, Windows
   :synopsis: Extends a cim.v1.activity.simulation_run document.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""



def get_extenders():
    """Returns set of extension functions.

    """
    return (
        _set_experiment_conformances,
        _set_responsible_parties
        )


def _set_experiment_conformances(ctx):
    """Sets conformance links.

    """
    ctx.ext.conformances = []
    for conformance in ctx.doc.conformances:
    	if len(conformance.requirements):
    		for requirement in conformance.requirements:
    			ctx.ext.conformances.append(requirement.name)
    ctx.ext.conformances = ", ".join(sorted(ctx.ext.conformances))


def _set_responsible_parties(ctx):
    """Sets sorted contacts list.

    """
    ctx.doc.responsible_parties = \
        sorted(ctx.doc.responsible_parties, key=lambda rp: rp.role)




# -*- coding: utf-8 -*-
"""
.. module:: cim.__init__.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: The cim package initializer.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using pyesdoc @ 2013-04-17 13:14:13.903808.

"""
import v1
import v2
import constants


# Supported version of this ontology.
VERSIONS = (v1, v2)



# -*- coding: utf-8 -*-

"""
.. module:: cim.v1.typeset_for_activity_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: The set of types of the cim.v1.activity package.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
import abc
import datetime
import uuid

import typeset_for_shared_package as shared



class Activity(object):
    """An abstract class within the cim v1 type system.

    An abstract class used as the parent of MeasurementCampaigns, Projects, Experiments, and NumericalActivities.

    """
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        """Instance constructor.

        """
        super(Activity, self).__init__()

        self.funding_sources = []                         # unicode (0.N)
        self.projects = []                                # activity.ProjectType (0.N)
        self.rationales = []                              # unicode (0.N)
        self.responsible_parties = []                     # shared.ResponsibleParty (0.N)


class Conformance(object):
    """A concrete class within the cim v1 type system.

    A conformance class maps how a configured model component met a specific numerical requirement.  For example, for a double CO2 boundary condition, a model component might read a CO2 dataset in which CO2 has been doubled, or it might modify a parameterisation (presumably with a factor of two somewhere).  So, the conformance links a requirement to a DataSource (which can be either an actual DataObject or a property of a model component).  In some cases a model/simulation may _naturally_ conform to a requirement.  In this case there would be no reference to a DataSource but the conformant attribute would be true.  If something is purpopsefully non-conformant then the conformant attribute would be false.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Conformance, self).__init__()

        self.description = None                           # unicode (0.1)
        self.frequency = None                             # activity.FrequencyType (0.1)
        self.is_conformant = None                         # bool (1.1)
        self.requirements = []                            # activity.NumericalRequirement (1.N)
        self.sources = []                                 # shared.DataSource (0.N)
        self.type = None                                  # activity.ConformanceType (0.1)


class ExperimentRelationship(shared.Relationship):
    """A concrete class within the cim v1 type system.

    Contains a set of relationship types specific to a experiment document that can be used to describe its genealogy.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(ExperimentRelationship, self).__init__()

        self.target = None                                # activity.ExperimentRelationshipTarget (1.1)
        self.type = None                                  # activity.ExperimentRelationshipType (1.1)


class ExperimentRelationshipTarget(object):
    """A concrete class within the cim v1 type system.

    Creates and returns instance of experiment_relationship_target class.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(ExperimentRelationshipTarget, self).__init__()

        self.numerical_experiment = None                  # activity.NumericalExperiment (0.1)


class NumericalRequirement(object):
    """An abstract class within the cim v1 type system.

    A description of the requirements of particular experiments.  Numerical Requirements can be initial conditions, boundary conditions, or physical modificiations.

    """
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        """Instance constructor.

        """
        super(NumericalRequirement, self).__init__()

        self.description = None                           # unicode (0.1)
        self.id = None                                    # unicode (0.1)
        self.name = None                                  # unicode (1.1)
        self.options = []                                 # activity.NumericalRequirementOption (0.N)
        self.requirement_type = None                      # unicode (1.1)
        self.source = None                                # shared.DataSource (0.1)


class NumericalRequirementOption(object):
    """A concrete class within the cim v1 type system.

    A NumericalRequirement that is being used as a set of related requirements; For example if a requirement is to use 1 of 3 boundary conditions, then that "parent" requirement would have three "child" RequirmentOptions (each of one with the XOR optionRelationship).

    """
    def __init__(self):
        """Instance constructor.

        """
        super(NumericalRequirementOption, self).__init__()

        self.description = None                           # unicode (0.1)
        self.id = None                                    # unicode (0.1)
        self.name = None                                  # unicode (1.1)
        self.relationship = None                          # unicode (0.1)
        self.sub_options = []                             # activity.NumericalRequirementOption (0.N)


class SimulationRelationship(shared.Relationship):
    """A concrete class within the cim v1 type system.

    Contains a set of relationship types specific to a simulation document that can be used to describe its genealogy.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(SimulationRelationship, self).__init__()

        self.target = None                                # activity.SimulationRelationshipTarget (1.1)
        self.type = None                                  # activity.SimulationRelationshipType (1.1)


class SimulationRelationshipTarget(object):
    """A concrete class within the cim v1 type system.

    Creates and returns instance of simulation_relationship_target class.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(SimulationRelationshipTarget, self).__init__()

        self.simulation = None                            # shared.DocReference (0.1)
        self.target = None                                # activity.SimulationType (0.1)


class BoundaryCondition(NumericalRequirement):
    """A concrete class within the cim v1 type system.

    A boundary condition is a numerical requirement which looks like a variable imposed on the model evolution (i.e. it might - or might not - evolve with time, but is seen by the model at various times during its evolution) as opposed to an initial condition (at model time zero).

    """
    def __init__(self):
        """Instance constructor.

        """
        super(BoundaryCondition, self).__init__()

        self.requirement_type = unicode("boundaryCondition")


class Experiment(Activity):
    """An abstract class within the cim v1 type system.

    An experiment might be an activity which is both observational and numerical in focus, for example, a measurement campaign and numerical experiments for an alpine experiment.  It is a place for the scientific description of the reason why an experiment was made.

    """
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        """Instance constructor.

        """
        super(Experiment, self).__init__()

        self.generates = []                               # unicode (0.N)
        self.measurement_campaigns = []                   # activity.MeasurementCampaign (0.N)
        self.requires = []                                # activity.NumericalActivity (0.N)
        self.supports = []                                # unicode (0.N)


class InitialCondition(NumericalRequirement):
    """A concrete class within the cim v1 type system.

    An initial condition is a numerical requirement on a model prognostic variable value at time zero.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(InitialCondition, self).__init__()

        self.requirement_type = unicode("initialCondition")


class LateralBoundaryCondition(NumericalRequirement):
    """A concrete class within the cim v1 type system.

    A boundary condition is a numerical requirement which looks like a variable imposed on the model evolution (i.e. it might - or might not - evolve with time, but is seen by the model at various times during its evolution) as opposed to an initial condition (at model time zero).

    """
    def __init__(self):
        """Instance constructor.

        """
        super(LateralBoundaryCondition, self).__init__()

        self.requirement_type = unicode("lateralBoundaryCondition")


class MeasurementCampaign(Activity):
    """A concrete class within the cim v1 type system.

    Creates and returns instance of measurement_campaign class.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(MeasurementCampaign, self).__init__()

        self.duration = None                              # int (1.1)


class NumericalActivity(Activity):
    """An abstract class within the cim v1 type system.

    Creates and returns instance of numerical_activity class.

    """
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        """Instance constructor.

        """
        super(NumericalActivity, self).__init__()

        self.description = None                           # unicode (0.1)
        self.long_name = None                             # unicode (0.1)
        self.short_name = None                            # unicode (1.1)
        self.supports = []                                # activity.Experiment (0.N)


class NumericalExperiment(Experiment):
    """A concrete class within the cim v1 type system.

    A numerical experiment may be generated by an experiment, in which case it is inSupportOf the experiment. But a numerical experiment may also exist as an activity in its own right (as it might be if it were needed for a MIP). Examples: AR4 individual experiments, AR5 individual experiments, RAPID THC experiments etc.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(NumericalExperiment, self).__init__()

        self.description = None                           # unicode (0.1)
        self.experiment_id = None                         # unicode (0.1)
        self.long_name = None                             # unicode (0.1)
        self.meta = shared.DocMetaInfo()                  # shared.DocMetaInfo (1.1)
        self.requirements = []                            # activity.NumericalRequirement (0.N)
        self.short_name = None                            # unicode (1.1)


class OutputRequirement(NumericalRequirement):
    """A concrete class within the cim v1 type system.

    Contains a set of relationship types specific to a simulation document that can be used to describe its genealogy.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(OutputRequirement, self).__init__()

        self.requirement_type = unicode("outputRequirement")


class PhysicalModification(Conformance):
    """A concrete class within the cim v1 type system.

    Physical modification is the implementation of a boundary condition numerical requirement that is achieved within the model code rather than from some external source file. It  might include, for example,  a specific rate constant within a chemical reaction, or coefficient value(s) in a parameterisation.  For example, one might require a numerical experiment where specific chemical reactions were turned off - e.g. no heterogeneous chemistry.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(PhysicalModification, self).__init__()



class Simulation(NumericalActivity):
    """An abstract class within the cim v1 type system.

    A simulation is the implementation of a numerical experiment.  A simulation can be made up of "child" simulations aggregated together to form a simulation composite.  The parent simulation can be made up of whole or partial child simulations, the simulation attributes need to be able to capture this.

    """
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        """Instance constructor.

        """
        super(Simulation, self).__init__()

        self.authors = None                               # unicode (0.1)
        self.calendar = None                              # shared.Calendar (1.1)
        self.conformances = []                            # activity.Conformance (0.N)
        self.control_simulation = None                    # activity.Simulation (0.1)
        self.deployments = []                             # software.Deployment (0.N)
        self.inputs = []                                  # software.Coupling (0.N)
        self.outputs = []                                 # data.DataObject (0.N)
        self.restarts = []                                # data.DataObject (0.N)
        self.simulation_id = None                         # unicode (0.1)
        self.spinup_date_range = None                     # shared.ClosedDateRange (0.1)
        self.spinup_simulation = None                     # activity.Simulation (0.1)


class SimulationComposite(Simulation):
    """A concrete class within the cim v1 type system.

    A SimulationComposite is an aggregation of Simulations. With the aggreation connector between Simulation and SimulationComposite(SC) the SC can be made up of both SimulationRuns and SCs. The SimulationComposite is the new name for the concept of SimulationCollection: A simulation can be made up of "child" simulations aggregated together to form a "simulation composite".  The "parent" simulation can be made up of whole or partial child simulations and the SimulationComposite attributes need to be able to capture this.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(SimulationComposite, self).__init__()

        self.child = []                                   # activity.Simulation (0.N)
        self.date_range = None                            # shared.DateRange (1.1)
        self.meta = shared.DocMetaInfo()                  # shared.DocMetaInfo (1.1)
        self.rank = None                                  # int (1.1)


class SimulationRun(Simulation):
    """A concrete class within the cim v1 type system.

    A SimulationRun is, as the name implies, one single model run. A SimulationRun is a Simulation. There is a one to one association between SimulationRun and (a top-level) SoftwarePackage::ModelComponent.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(SimulationRun, self).__init__()

        self.date_range = None                            # shared.DateRange (1.1)
        self.meta = shared.DocMetaInfo()                  # shared.DocMetaInfo (1.1)
        self.model = None                                 # software.ModelComponent (0.1)


class SpatioTemporalConstraint(NumericalRequirement):
    """A concrete class within the cim v1 type system.

    Contains a set of relationship types specific to a simulation document that can be used to describe its genealogy.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(SpatioTemporalConstraint, self).__init__()

        self.date_range = None                            # shared.DateRange (0.1)
        self.spatial_resolution = None                    # activity.ResolutionType (0.1)
        self.requirement_type = unicode("spatioTemporalConstraint")


class DownscalingSimulation(NumericalActivity):
    """An abstract class within the cim v1 type system.

    A simulation is the implementation of a numerical experiment.  A simulation can be made up of "child" simulations aggregated together to form a simulation composite.  The parent simulation can be made up of whole or partial child simulations, the simulation attributes need to be able to capture this.

    """
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        """Instance constructor.

        """
        super(DownscalingSimulation, self).__init__()

        self.calendar = None                              # shared.Calendar (1.1)
        self.downscaled_from = None                       # shared.DataSource (1.1)
        self.downscaling_id = None                        # unicode (0.1)
        self.downscaling_type = None                      # activity.DownscalingType (0.1)
        self.inputs = []                                  # software.Coupling (0.N)
        self.meta = shared.DocMetaInfo()                  # shared.DocMetaInfo (1.1)
        self.outputs = []                                 # data.DataObject (0.N)


class Ensemble(NumericalActivity):
    """A concrete class within the cim v1 type system.

    An ensemble is made up of two or more simulations which are to be compared against each other to create ensemble statistics. Ensemble members can differ in terms of initial conditions, physical parameterisation and the model used. An ensemble bundles together sets of ensembleMembers, all of which reference the same Simulation(Run) and include one or more changes.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Ensemble, self).__init__()

        self.members = []                                 # activity.EnsembleMember (1.N)
        self.meta = shared.DocMetaInfo()                  # shared.DocMetaInfo (1.1)
        self.outputs = []                                 # shared.DataSource (0.N)
        self.types = []                                   # activity.EnsembleType (1.N)


class EnsembleMember(NumericalActivity):
    """A concrete class within the cim v1 type system.

    A simulation is the implementation of a numerical experiment.  A simulation can be made up of "child" simulations aggregated together to form a "simulation composite".  The "parent" simulation can be made up of whole or partial child simulations, the simulation attributes need to be able to capture this.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(EnsembleMember, self).__init__()

        self.ensemble = None                              # activity.Ensemble (0.1)
        self.ensemble_ids = []                            # shared.StandardName (0.N)
        self.simulation = None                            # activity.Simulation (0.1)


class ConformanceType(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of conformance_type enum.
    """
    is_open = False
    members = [
        "not-xxx",
        "not conformant",
        "standard config",
        "via inputs",
        "via model mods",
        "combination"
        ]
    descriptions = [
        "None",
        "Describes a simulation that is purpefully non-conformant to an experimental requirement.",
        "Describes a simulation that is naturally conformant to an experimental requirement.",
        "Describes a simulation that conforms to an experimental requirement by using particular inputs.",
        "Describes a simulation that conforms to an experimental requirement by changing the configuration of the software model implementing that simulation.",
        "Describes a simulation that conforms to an experimental requirement by using more than one method."
        ]


class DownscalingType(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of downscaling_type enum.
    """
    is_open = False
    members = [
        "statistical",
        "dynamic"
        ]
    descriptions = [
        "None",
        "None"
        ]


class EnsembleType(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of ensemble_type enum.
    """
    is_open = True
    members = []
    descriptions = []


class ExperimentRelationshipType(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of experiment_relationship_type enum.
    """
    is_open = False
    members = [
        "previousRealisation",
        "continuationOf",
        "controlExperiment",
        "higherResolutionVersionOf",
        "lowerResolutionVersionOf",
        "increaseEnsembleOf",
        "modifiedInputMethodOf",
        "shorterVersionOf",
        "extensionOf"
        ]
    descriptions = [
        "None",
        "None",
        "None",
        "None",
        "None",
        "None",
        "None",
        "None",
        "None"
        ]


class FrequencyType(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of frequency_type enum.
    """
    is_open = True
    members = []
    descriptions = []


class ProjectType(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of project_type enum.
    """
    is_open = True
    members = []
    descriptions = []


class ResolutionType(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of resolution_type enum.
    """
    is_open = True
    members = []
    descriptions = []


class SimulationRelationshipType(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of simulation_relationship_type enum.
    """
    is_open = False
    members = [
        "extensionOf",
        "responseTo",
        "continuationOf",
        "previousSimulation",
        "higherResolutionVersionOf",
        "lowerResolutionVersionOf",
        "fixedVersionOf",
        "followingSimulation"
        ]
    descriptions = [
        "None",
        "None",
        "None",
        "None",
        "None",
        "None",
        "None",
        "None"
        ]


class SimulationType(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of simulation_type enum.
    """
    is_open = False
    members = [
        "simulationRun",
        "assimilation",
        "simulationComposite"
        ]
    descriptions = [
        "None",
        "None",
        "None"
        ]





# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.drq.utils.py
   :copyright: Copyright "February 27, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates library utility functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import os
import re
import xml.etree.ElementTree as ET

from pyesdoc.drq import constants



class _State(object):
    """Encpasulates mutable module state.

    """
    # Flag indicating whether state has been initialized.
    _is_initialized = False


def initialize():
    """Loads & parses dreq2Defn.xml & dreq.xml files.

    """
    if _State._is_initialized:
        return

    from pyesdoc.drq.definition import initializer
    initializer.execute()

    from pyesdoc.drq.content import initializer
    initializer.execute()

    _State._is_initialized = True


def init_from_xml(
    instance,
    elem,
    names,
    convertors=None
    ):
    """Initialises set of class instance attributes from an xml element.

    :param object instance: Class instance to be initialized.
    :param xml.etree.Element elem: XML element used to hydrate class instance.
    :param list names: Set of class attribute names.
    :param dict convertors: Map of value convertors keyed by attribute name.

    """
    names = [(n, get_label(n)) for n in names]
    for name_xml, name_py in names:
        # Set value.
        val = None
        try:
            elem.get
        except AttributeError:
            child_elem = elem.find('./{}'.format(name_xml))
            if child_elem:
                val = child_elem.text
        else:
            val = elem.get(name_xml)

        # Set value to None.
        if val and unicode(val).strip().lower() in {u'', u'none'}:
            val = None

        # Convert value.
        if convertors and name_py in convertors:
            try:
                val = convertors[name_py](val)
            except TypeError:
                if val is None:
                    val = convertors[name_py]()

        # Set instance attribute using pythonic name.
        setattr(instance, name_py, val)


def load_xml(fpath):
    """Loads an XML document from file system stripping out namespaces as it does so.

    """
    if not os.path.exists(fpath):
      raise IOError("Invalid data request file path: {}".format(fpath))

    iterator = ET.iterparse(fpath)
    for _, elem in iterator:
        if '}' in elem.tag:
            elem.tag = elem.tag.split('}', 1)[1]

    return iterator.root


def get_fpath(fname, subfolder):
    """Returns path to a data request file.

    """
    fpath = os.path.dirname(os.path.abspath(__file__))
    fpath = os.path.join(fpath, subfolder)
    fpath = os.path.join(fpath, fname)

    return fpath


def get_label(label):
    """Returns a reformatted label.

    """
    if label in constants.LABEL_MAP:
        return constants.LABEL_MAP[label]
    elif label != label.lower():
        return str_to_underscore_case(label)
    return label


_CACHE = set()

def get_link_labels(label):
    """Returns a reformatted link label.

    """
    try:
        return constants.LINK_MAP[label]
    except KeyError:
        if label not in _CACHE:
            _CACHE.add(label)
        return label, label


def str_to_underscore_case(target):
    """Helper function to convert a from camel case string to an underscore case string.

    :param target: A string for conversion.
    :type target: str

    :returns: A string converted to underscore case, e.g. account_number.
    :rtype: str

    """
    if target is None or not len(target):
        return ''

    r = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', target)
    r = re.sub('([a-z0-9])([A-Z])', r'\1_\2', r)
    r = r.lower()

    return r


# -*- coding: utf-8 -*-

"""
.. module:: software_enums.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Set of CIM v1 software package enum definitions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""



def component_property_intent_type():
    """The direction that the associated component property is intended to be coupled: in, out, or inout.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ('in', None),
            ('out', None),
            ('inout', None),
        ],
    }


def connection_type():
    """The ConnectionType enumeration describes the mechanism of transport for a connection.

    """
    return {
        'type': 'enum',
        'is_open': True
    }


def coupling_framework_type():
    """Creates and returns instance of coupling_framework_type enum.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ('BFG', None),
            ('ESMF', None),
            ('OASIS', None),
        ],
    }


def model_component_type():
    """An enumeration of types of ModelComponent. This includes things like atmosphere & ocean models, radiation schemes, etc. CIM best-practice is to describe every component for which there is a named ComponentType as a separate component, even if it is not a separate unit of software (ie: even if it is embedded), instead of as a (set of) ModelParameters. This codelist is synonomous with "realm" for the purposes of CMIP5.

    """
    return {
        'type': 'enum',
        'is_open': True
    }


def spatial_regridding_dimension_type():
    """Creates and returns instance of spatial_regridding_dimension_type enum.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ('1D', None),
            ('2D', None),
            ('3D', None),
        ],
    }


def spatial_regridding_standard_method_type():
    """Creates and returns instance of spatial_regridding_standard_method_type enum.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ('linear', None),
            ('near-neighbour', None),
            ('cubic', None),
            ('conservative-first-order', None),
            ('conservative-second-order', None),
            ('conservative', None),
            ('non-conservative', None),
        ],
    }


def statistical_model_component_type():
    """An enumeration of types of ProcessorComponent.  This includes things like transformers and post-processors.

    """
    return {
        'type': 'enum',
        'is_open': True
    }


def time_mapping_type():
    """Enumerates the different ways that time can be mapped when transforming from one field to another.

    """
    return {
        'type': 'enum',
        'is_open': True
    }


def timing_units():
    """Creates and returns instance of timing_units enum.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ('seconds', None),
            ('minutes', None),
            ('hours', None),
            ('days', None),
            ('months', None),
            ('years', None),
            ('decades', None),
            ('centuries', None),
        ],
    }



# -*- coding: utf-8 -*-

"""
.. module:: cim.v2.typeset_for_shared_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: The set of types of the cim.v2.shared package.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
import abc
import datetime
import uuid




class Citation(object):
    """A concrete class within the cim v2 type system.

    An academic reference to published work.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Citation, self).__init__()

        self.abstract = None                              # unicode (0.1)
        self.citation_detail = None                       # unicode (0.1)
        self.collective_title = None                      # unicode (0.1)
        self.context = None                               # unicode (0.1)
        self.doi = None                                   # unicode (0.1)
        self.meta = DocMetaInfo()                         # shared.DocMetaInfo (1.1)
        self.title = None                                 # unicode (0.1)
        self.type = None                                  # unicode (0.1)
        self.url = None                                   # shared.OnlineResource (0.1)


class DocMetaInfo(object):
    """A concrete class within the cim v2 type system.

    Encapsulates document meta information used by es-doc machinery. Will not normally be
    populated by humans. May duplicate information held in 'visible' metadata.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DocMetaInfo, self).__init__()

        self.author = None                                # shared.Party (0.1)
        self.create_date = None                           # datetime.datetime (1.1)
        self.drs_keys = []                                # unicode (0.N)
        self.drs_path = None                              # unicode (0.1)
        self.external_ids = []                            # unicode (0.N)
        self.id = None                                    # unicode (1.1)
        self.institute = None                             # unicode (0.1)
        self.project = None                               # unicode (0.1)
        self.sort_key = None                              # unicode (0.1)
        self.source = None                                # unicode (1.1)
        self.source_key = None                            # unicode (0.1)
        self.sub_projects = []                            # unicode (0.N)
        self.type = None                                  # unicode (1.1)
        self.type_display_name = None                     # unicode (0.1)
        self.type_sort_key = None                         # unicode (0.1)
        self.update_date = None                           # datetime.datetime (0.1)
        self.version = None                               # int (1.1)


class DocReference(object):
    """A concrete class within the cim v2 type system.

    A reference to another cim entity.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DocReference, self).__init__()

        self.canonical_name = None                        # unicode (0.1)
        self.constraint_vocabulary = None                 # unicode (0.1)
        self.context = None                               # unicode (0.1)
        self.description = None                           # unicode (0.1)
        self.external_id = None                           # unicode (0.1)
        self.further_info = None                          # unicode (0.1)
        self.id = None                                    # unicode (0.1)
        self.institute = None                             # unicode (0.1)
        self.linkage = None                               # unicode (0.1)
        self.name = None                                  # unicode (0.1)
        self.protocol = None                              # unicode (0.1)
        self.relationship = None                          # unicode (0.1)
        self.type = None                                  # unicode (0.1)
        self.url = None                                   # unicode (0.1)
        self.version = None                               # int (0.1)


class ExtraAttribute(object):
    """A concrete class within the cim v2 type system.

    An extra attribute with key and value needed to encode further information
    not in the CIM2 domain model or specialisation. Typical use case: in parsing
    data and encoding attributes found in data.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(ExtraAttribute, self).__init__()

        self.doc = None                                   # unicode (0.1)
        self.key = None                                   # unicode (1.1)
        self.type = None                                  # unicode (0.1)
        self.value = None                                 # unicode (1.1)


    @property
    def __str__(self):
	    """Instrance string representation.

	    """
	    return "{}:{}".format(self.key, self.value)


class OnlineResource(object):
    """A concrete class within the cim v2 type system.

    A minimal approximation of ISO19115 CI_ONLINERESOURCE, provides a link and details
    of how to use that link.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(OnlineResource, self).__init__()

        self.description = None                           # unicode (0.1)
        self.linkage = None                               # unicode (1.1)
        self.name = None                                  # unicode (1.1)
        self.protocol = None                              # unicode (0.1)


class Party(object):
    """A concrete class within the cim v2 type system.

    Implements minimal material for an ISO19115-1 (2014) compliant party.
    For our purposes this is a much better animal than the previous responsibleParty 
    which munged roles together with people. Note we have collapsed CI_Contact,
    CI_Individual and CI_Organisation as well as the abstract CI_Party.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Party, self).__init__()

        self.address = None                               # unicode (0.1)
        self.email = None                                 # unicode (0.1)
        self.is_organisation = None                       # bool (0.1)
        self.meta = DocMetaInfo()                         # shared.DocMetaInfo (1.1)
        self.name = None                                  # unicode (0.1)
        self.orcid_id = None                              # unicode (0.1)
        self.url = None                                   # shared.OnlineResource (0.1)


class QualityReview(object):
    """A concrete class within the cim v2 type system.

    Assertions as to the completeness and quality of a document.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(QualityReview, self).__init__()

        self.date = None                                  # unicode (1.1)
        self.meta = DocMetaInfo()                         # shared.DocMetaInfo (1.1)
        self.metadata_reviewer = None                     # shared.Party (1.1)
        self.quality_description = None                   # unicode (1.1)
        self.quality_status = None                        # shared.QualityStatus (0.1)
        self.target_document = None                       # shared.DocReference (1.1)


class Responsibility(object):
    """A concrete class within the cim v2 type system.

    Implements the ISO19115-1 (2014) CI_Responsibility (which replaces
    responsibleParty). Combines a person and their role in doing something.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Responsibility, self).__init__()

        self.parties = []                                 # shared.Party (1.N)
        self.role = None                                  # shared.RoleCode (1.1)
        self.when = None                                  # time.TimePeriod (0.1)


    @property
    def __str__(self):
	    """Instrance string representation.

	    """
	    return "{}:{}".format(self.role, self.parties)


class TextBlob(object):
    """A concrete class within the cim v2 type system.

    Provides a text class which supports plaintext, html, and
    friends (or will do).

    """
    def __init__(self):
        """Instance constructor.

        """
        super(TextBlob, self).__init__()

        self.content = None                               # unicode (1.1)
        self.encoding = None                              # shared.TextBlobEncoding (1.1)


    @property
    def __str__(self):
	    """Instrance string representation.

	    """
	    return "{}".format(self.content)


class NilReason(object):
    """An enumeration within the cim v2 type system.

    Provides an enumeration of possible reasons why a property has not been defined
    Based on GML nilReason as discussed here: https://www.seegrid.csiro.au/wiki/AppSchemas/NilValues.
    """
    is_open = False
    members = [
        "nil:inapplicable",
        "nil:missing",
        "nil:template",
        "nil:unknown",
        "nil:withheld"
        ]
    descriptions = [
        "There is no value",
        "The correct value is not available. Furthermore, a correct value may not exist",
        "The value will be available later",
        "The correct value is not known at this time. However, a correct value probably exists",
        "The value is not divulged"
        ]


class QualityStatus(object):
    """An enumeration within the cim v2 type system.

    Assertion as to the review status of document.
    """
    is_open = False
    members = [
        "incomplete",
        "finalised",
        "under_review",
        "reviewed"
        ]
    descriptions = [
        "Currently being worked on",
        "Author has completed document, prior to review",
        "Document is being reviewed",
        "Document has been formally reviewed and assessed as complete and accurate"
        ]


class RoleCode(object):
    """An enumeration within the cim v2 type system.

    Responsibility role codes: roles that a party may play in delivering a responsibility.
    """
    is_open = False
    members = [
        "Principal Investigator",
        "originator",
        "author",
        "collaborator",
        "publisher",
        "owner",
        "processor",
        "distributor",
        "sponsor",
        "user",
        "point of contact",
        "resource provider",
        "custodian",
        "metadata_reviewer",
        "metadata_author"
        ]
    descriptions = [
        "Key party responsible for the existence of the resource",
        "Original source for the resource if obtained from elsewhere",
        "Party who created (or co-created) resource",
        "Contributor to the production of the resource",
        "Party who published the resource",
        "Party with legal ownership of the resource",
        "Party who has taken part in the workflow that resulted in this resource",
        "Party who distributes the resource",
        "Party who has invested in the production of the resource",
        "Party who uses the resource",
        "Party who can be contacted for acquiring knowledge about or acquisition of the resource",
        "Party that supplies the resource",
        "Party that accepts accountability and responsibility for the source resource",
        "Party who carried out an independent review of (this) documentation",
        "Party who created (this) documentation"
        ]


class TextBlobEncoding(object):
    """An enumeration within the cim v2 type system.

    Types of text understood by the CIM notebook. Currently only
    plaintext, but we expect safe HTML to be supported as soon as practicable.
    """
    is_open = False
    members = [
        "ascii"
        ]
    descriptions = [
        "Normal plain text"
        ]


